from __future__ import annotations

import logging
from functools import partial
from pathlib import Path
from typing import Any, Callable, TYPE_CHECKING, Iterable, Tuple, TypeVar, overload
import warnings
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Signal, Qt

import numpy as np
import pandas as pd
from collections_undo import fmt, arguments

from ._item_model import AbstractDataFrameModel
from ._line_edit import (
    QHorizontalHeaderLineEdit,
    QVerticalHeaderLineEdit,
    QCellLiteralEdit,
)
from tabulous._sort_filter_proxy import SortFilterProxy
from tabulous._dtype import isna
from tabulous._qt._undo import QtUndoManager, fmt_slice
from tabulous._qt._svg import QColoredSVGIcon
from tabulous._keymap import QtKeys, QtKeyMap
from tabulous._qt._action_registry import QActionRegistry
from tabulous.types import ProxyType, ItemInfo, HeaderInfo, EvalInfo
from tabulous.exceptions import (
    SelectionRangeError,
    TableImmutableError,
    UnreachableError,
)
from tabulous._pd_index import as_not_ranged, as_constructor, is_ranged

if TYPE_CHECKING:
    from ._delegate import TableItemDelegate
    from ._side_area import QTableSideArea
    from ._enhanced_table import _QTableViewEnhanced
    from tabulous._qt._table_stack import QTabbedTableStack
    from tabulous._qt._mainwindow import _QtMainWidgetBase
    from tabulous.types import SelectionType, _Sliceable
    from tabulous._psygnal import InCellRangedSlot

logger = logging.getLogger("tabulous")
ICON_DIR = Path(__file__).parent.parent.parent / "_icons"

_SplitterStyle = """
QSplitter::handle {
    background-color: gray;
    border: 0px;
    height: 2px;
    margin-left: 5px;
    margin-right: 5px;
    border-radius: 1px;
}
"""


class QTableHandle(QtW.QSplitterHandle):
    """The handle widget used to resize the side area."""

    def __init__(self, o: Qt.Orientation, parent: QtW.QSplitter) -> None:
        super().__init__(o, parent)
        self._sizes = parent.sizes()

    def mouseDoubleClickEvent(self, a0: QtGui.QMouseEvent) -> None:
        """Collapse/expand side area."""
        parent = self.splitter()
        sizes = parent.sizes()
        if sizes[1] == 0:
            parent.setSizes(self._sizes)
        else:
            self._sizes = sizes
            parent.setSizes([1, 0])

        return super().mouseDoubleClickEvent(a0)


class QBaseTable(QtW.QSplitter, QActionRegistry[Tuple[int, int]]):
    """
    The base widget for a table.

    Abstract Methods
    ----------------
    def createQTableView(self) -> None: ...
    def getDataFrame(self) -> pd.DataFrame: ...
    def setDataFrame(self) -> None: ...
    def createModel(self) -> AbstractDataFrameModel: ...
    def tableSlice(self) -> pd.DataFrame: ...
    """

    selectionChangedSignal = Signal()
    _DEFAULT_EDITABLE = False
    _mgr = QtUndoManager()
    _keymap = QtKeyMap()

    def __init__(
        self, parent: QtW.QWidget | None = None, data: pd.DataFrame | None = None
    ):
        QtW.QSplitter.__init__(self, parent)
        QActionRegistry.__init__(self)

        self._proxy = SortFilterProxy()
        self._filtered_index: pd.Index | None = None
        self._filtered_columns: pd.Index | None = None
        self.setContentsMargins(0, 0, 0, 0)

        self.createQTableView()
        self.createModel()
        self.setDataFrame(data)

        self._qtable_view.selectionChangedSignal.connect(
            self.selectionChangedSignal.emit
        )

        self._side_area: QTableSideArea = None
        self.model()._editable = self._DEFAULT_EDITABLE
        self.setStyleSheet(_SplitterStyle)

        self._qtable_view.rightClickedSignal.connect(self.showContextMenu)
        self._edited = False  # data loaded from a file is edited or not

    def needSave(self) -> bool:
        """Return True if the table is edited."""
        return self._edited

    def setOrientation(self, a0: Qt.Orientation) -> None:
        """Set table orientation and the side area orientation."""
        a1 = 1 - a0  # opposite orientation
        super().setOrientation(a0)
        if self._side_area is not None:
            self._side_area.setOrientation(a1)
        return None

    def createHandle(self) -> QTableHandle:
        """Create custom handle."""
        return QTableHandle(Qt.Orientation.Horizontal, self)

    def showContextMenu(self, pos: QtCore.QPoint) -> None:
        """Show contextmenu at the given position."""
        index = self._qtable_view.indexAt(pos)
        if not index.isValid():
            return None

        row, col = index.row(), index.column()
        sel_model = self._qtable_view._selection_model
        highlight_model = self._qtable_view._highlight_model

        if sel_model._ctrl_on:
            # if Ctrl is on, select the highlight under the cursor.
            idx, rng = highlight_model.range_under_index(row, col)
            if rng is not None:
                if len(highlight_model._selected_indices) == 0:
                    self.setSelections([rng])
                else:
                    self.setSelections(self.selections() + [rng])
                highlight_model.add_selection(idx)
            else:
                self.setSelections([(row, col)])
                highlight_model.select([])
        else:
            idx, rng = sel_model.range_under_index(row, col)
            if rng is not None:
                sel_model.reorder_to_last(idx)
                self.update()
            else:
                sel_model.current_index = (row, col)
                self.setSelections([(row, col)])
            highlight_model.select([])

        return self.execContextMenu(
            self._qtable_view.viewport().mapToGlobal(pos), (row, col)
        )

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        raise NotImplementedError()

    @property
    def _central_widget(self) -> QtW.QWidget:
        return self._qtable_view

    def createQTableView(self) -> None:
        """Create QTableView."""
        raise NotImplementedError()

    def getDataFrame(self) -> pd.DataFrame:
        raise NotImplementedError()

    # fmt: off
    @overload
    def _get_sub_frame(self, columns: list[str]) -> pd.DataFrame: ...
    @overload
    def _get_sub_frame(self, columns: str) -> pd.Series: ...
    # fmt: on

    def _get_sub_frame(self, columns):
        return self.getDataFrame()[columns]

    def setDataFrame(self, df: pd.DataFrame) -> None:
        raise NotImplementedError()

    def createModel(self) -> AbstractDataFrameModel:
        raise NotImplementedError()

    def tableSlice(self) -> pd.DataFrame:
        raise NotImplementedError()

    def tableShape(self) -> tuple[int, int]:
        model = self._qtable_view.model()
        nr = model.rowCount()
        nc = model.columnCount()
        return (nr, nc)

    def dataShape(self) -> tuple[int, int]:
        """Shape of the visible data"""
        return self.tableShape()

    def dataShapeRaw(self) -> tuple[int, int]:
        return self.getDataFrame().shape

    def zoom(self) -> float:
        """Get current zoom factor."""
        return self._qtable_view.zoom()

    def setZoom(self, value: float) -> None:
        """Set zoom factor."""
        return self._qtable_view.setZoom(value)

    def itemDelegate(self) -> TableItemDelegate:
        return QtW.QTableView.itemDelegate(self._qtable_view)

    def model(self) -> AbstractDataFrameModel:
        return self._qtable_view.model()

    def setDataFrameValue(self, row: int, col: int, value: Any) -> None:
        raise TableImmutableError("Table is immutable.")

    def setLabeledData(self, row, col, value) -> None:
        raise TableImmutableError("Table is immutable.")

    def deleteValues(self) -> None:
        raise TableImmutableError("Table is immutable.")

    def setVerticalHeaderValue(self, index: int, value: Any) -> None:
        raise TableImmutableError("Table is immutable.")

    def setHorizontalHeaderValue(self, index: int, value: Any) -> None:
        raise TableImmutableError("Table is immutable.")

    def isEditable(self) -> bool:
        """Return the editability of the table."""
        return False

    def setEditable(self, editable: bool):
        """Set the editability of the table."""
        if editable:
            raise TableImmutableError("Table is immutable.")

    def assignColumns(self, serieses: list[pd.Series]) -> None:
        raise TableImmutableError("Table is immutable.")

    def convertValue(self, c: int, value: Any) -> Any:
        """Convert value before updating DataFrame."""
        return value

    def _get_converter(self, c: int) -> Callable[[Any, Any], Any]:
        if 0 <= c < len(self._filtered_columns):
            colname = self._filtered_columns[c]
            if validator := self.model()._validator.get(colname, None):
                return partial(
                    _convert_value, validator=validator, converter=self.convertValue
                )
        return self.convertValue

    def dataShown(self, parse: bool = False) -> pd.DataFrame:
        """Return the shown dataframe (consider filter)."""
        return self.model().df

    def connectSelectionChangedSignal(self, slot):
        self.selectionChangedSignal.connect(slot)
        return slot

    def selections(self, map: bool = False) -> SelectionType:
        """Get list of selections as slicable tuples"""
        ranges = self._qtable_view._selection_model.as_ranges()
        if not map:
            return ranges
        return [
            (self._proxy.get_source_slice(r, force_single_row=True), c)
            for r, c in ranges
        ]

    def setSelections(self, selections: SelectionType):
        """Set list of selections."""
        qtable = self._qtable_view
        qtable.clear_selections()
        nr, nc = self.tableShape()
        qtable.set_selections(_normalize_selections(selections, nr, nc))
        self.update()
        return None

    def highlights(self) -> SelectionType:
        """Get list of selections as slicable tuples"""
        return self._qtable_view._highlight_model.as_ranges()

    @_mgr.interface
    def setHighlights(self, selections: SelectionType):
        """Set list of selections."""
        qtable = self._qtable_view
        qtable.clear_highlights()
        nr, nc = self.tableShape()
        qtable.set_highlights(_normalize_selections(selections, nr, nc))
        self.update()
        return None

    @setHighlights.server
    def setHighlights(self, selections: SelectionType):
        return arguments(self.highlights())

    def _selected_dataframe(self) -> pd.DataFrame:
        """Currently selected cells as a data frame."""
        selections = self.selections()
        if len(selections) == 0:
            return
        r_ranges = set()
        c_ranges = set()
        for rsel, csel in selections:
            r_ranges.add((rsel.start, rsel.stop))
            c_ranges.add((csel.start, csel.stop))

        nr = len(r_ranges)
        nc = len(c_ranges)
        if nr > 1 and nc > 1:
            raise SelectionRangeError("Cannot copy selected range.")
        data = self.dataShown()
        if nr == 1:
            axis = 1
        else:
            axis = 0
        ref = pd.concat([data.iloc[sel] for sel in selections], axis=axis)
        return ref

    def copyToClipboard(self, headers: bool = True, sep: str = "\t") -> None:
        """Copy currently selected cells to clipboard."""
        ref = self._selected_dataframe()
        ref.to_clipboard(index=headers, header=headers, sep=sep)
        return None

    def _copy_as_literal(self):
        """Copy the selected cells as evaluable string."""
        sels = self.selections()
        if len(sels) == 0:
            raise SelectionRangeError("No selection found.")
        table_stack = self.tableStack()
        if table_stack is None:
            raise ValueError("Table is not in a table stack.")
        i = table_stack.currentIndex()
        viewer = table_stack.parentWidget()._table_viewer
        name = viewer.tables[i].name
        _sl = _selection_to_literal(sels[-1])
        return _to_clipboard(f"viewer.tables[{name!r}].data.iloc{_sl}", excel=False)

    def _copy_as_formated(self, format: str = "markdown"):
        """Copy the selected cells as markdown format."""
        ref = self._selected_dataframe()
        index = not is_ranged(ref.index)
        if format == "markdown":
            s = _tabulate(ref, tablefmt="github", index=index)
        elif format == "latex":
            s = ref.to_latex(index=index)
        elif format == "html":
            s = ref.to_html(index=index)
        elif format == "rst":
            s = _tabulate(ref, tablefmt="rst", index=index)
        elif format == "grid":
            s = _dataframe_to_grid_table(ref)
        else:
            raise ValueError(f"Unknown format: {format!r}")
        return _to_clipboard(s, excel=False)

    def _copy_as_new_table(self, type_: str = "same"):
        """Copy the selected range to a new table."""
        sels = self.selections()
        if len(sels) != 1:
            raise SelectionRangeError("Inappropriate selection range.")

        table_stack = self.tableStack()
        if table_stack is None:
            raise ValueError("Table is not in a table stack.")
        i = table_stack.currentIndex()
        viewer = table_stack.parentWidget()._table_viewer
        name = viewer.tables[i].name

        df_cropped = self.getDataFrame().iloc[sels[0]]
        if type_ == "table":
            viewer.add_table(df_cropped, name=f"{name}-cropped", copy=False)
        elif type_ == "spreadsheet":
            viewer.add_spreadsheet(df_cropped, name=f"{name}-cropped", copy=False)
        elif type_ == "same":
            if viewer.tables[i].table_type == "SpreadSheet":
                viewer.add_spreadsheet(df_cropped, name=f"{name}-cropped", copy=False)
            else:
                viewer.add_table(df_cropped, name=f"{name}-cropped", copy=False)
        else:
            raise RuntimeError(f"Unknown type: {type_!r}")

    def pasteFromClipBoard(self):
        raise TableImmutableError("Table is immutable.")

    def readClipBoard(self, sep=r"\s+") -> pd.DataFrame:
        """Read clipboard data and return as pandas DataFrame."""
        return pd.read_clipboard(header=None, sep=sep)

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        if self._keymap.press_key(e):
            return None
        return super().keyPressEvent(e)

    def proxy(self) -> SortFilterProxy:
        """The sort/filter proxy object."""
        return self._proxy

    def setProxy(self, proxy: ProxyType, clear_header_widgets: bool = True):
        """Set sort/filter to the table view."""
        with self._mgr.merging(lambda cmds: f"table.proxy = {proxy!r}"):
            if clear_header_widgets:
                self.updateHorizontalHeaderWidget({})
            self._set_proxy(proxy)
        return None

    @_mgr.interface
    def _set_proxy(self, proxy: ProxyType):
        # NOTE: This method is also called when table needs initialization.
        self._proxy = SortFilterProxy(proxy)
        try:
            df_filt = self._apply_proxy()
        except Exception as e:
            # To avoid continuous error, set proxy to None.
            self._set_proxy(None)
            raise e

        prx = self._proxy
        proxy_type = prx.proxy_type
        if proxy_type == "none":
            icon = QtGui.QIcon()
        elif proxy_type == "filter":
            icon = QColoredSVGIcon.fromfile(ICON_DIR / "filter.svg")
        elif proxy_type == "sort":
            icon = QColoredSVGIcon.fromfile(ICON_DIR / "sort_table.svg")
        else:
            raise RuntimeError(f"Unknown proxy type: {proxy_type!r}")

        # update data
        self.model().df = df_filt
        self._filtered_index = df_filt.index
        self._filtered_columns = df_filt.columns

        # update filter icon
        if stack := self.tableStack():
            idx = stack.tableIndex(self)
            bg = self.palette().color(self.backgroundRole())
            whiteness = bg.red() + bg.green() + bg.blue()
            if not icon.isNull():
                if whiteness <= 128 * 3:
                    icon = icon.colored("#CCCCCC")
                else:
                    icon = icon.colored("#1E1E1E")

            stack.setTabIcon(idx, icon)
            if not icon.isNull():
                stack.setIconSize(QtCore.QSize(12, 12))

        return self.refreshTable()

    def _apply_proxy(self):
        data_sliced = self.tableSlice()
        return self._proxy.apply(data_sliced)

    @_set_proxy.server
    def _set_proxy(self, proxy: ProxyType):
        return arguments(self.proxy())

    @_set_proxy.set_formatter
    def _set_proxy_fmt(self, proxy: ProxyType):
        return f"table.proxy.set({proxy!r})"

    @_mgr.interface
    def setHorizontalHeaderWidget(self, idx: int, widget: QtW.QWidget | None):
        hheader = self._qtable_view.horizontalHeader()
        if widget is None:
            hheader.removeSectionWidget(idx)
        else:
            hheader.setSectionWidget(idx, widget)

    @setHorizontalHeaderWidget.server
    def setHorizontalHeaderWidget(self, idx: int, widget: QtW.QWidget | None):
        hheader = self._qtable_view.horizontalHeader()
        return arguments(idx, hheader.sectionWidget(idx))

    @_mgr.interface
    def updateHorizontalHeaderWidget(self, d: dict[int, QtW.QWidget] = {}):
        """Update horizontal header widgets."""
        hheader = self._qtable_view.horizontalHeader()
        hheader.removeSectionWidget()
        for idx, widget in d.items():
            hheader.setSectionWidget(idx, widget)

    @updateHorizontalHeaderWidget.server
    def updateHorizontalHeaderWidget(self, d: dict[int, QtW.QWidget]):
        hheader = self._qtable_view.horizontalHeader()
        return arguments(hheader._header_widgets.copy())

    @_mgr.interface
    def setForegroundColormap(self, name: str, colormap: Callable | None):
        """Set the colormap for the foreground."""
        if colormap is None:
            self.model()._foreground_colormap.pop(name, None)
        else:
            if not callable(colormap):
                raise TypeError("Cannot use non-callable objects as colormaps.")
            self.model()._foreground_colormap[name] = colormap
        self.refreshTable()
        return None

    @setForegroundColormap.server
    def setForegroundColormap(self, name: str, colormap: Callable):
        cmap = self.model()._foreground_colormap.get(name, None)
        return arguments(name, cmap)

    @_mgr.interface
    def setBackgroundColormap(self, name: str, colormap: Callable | None):
        """Set the colormap for the foreground."""
        if colormap is None:
            self.model()._background_colormap.pop(name, None)
        else:
            if not callable(colormap):
                raise TypeError("Cannot use non-callable objects as colormaps.")
            self.model()._background_colormap[name] = colormap
        self.refreshTable()
        return None

    @setBackgroundColormap.server
    def setBackgroundColormap(self, name: str, colormap: Callable):
        cmap = self.model()._background_colormap.get(name, None)
        return arguments(name, cmap)

    @_mgr.interface
    def setTextFormatter(self, name: str, fmt: Callable[[Any], str] | str):
        """Set a text formatter function to the column named `name`."""
        if fmt is None:
            self.model()._text_formatter.pop(name, None)
        else:
            if isinstance(fmt, str):
                fmt = fmt.format
            elif not callable(fmt):
                raise TypeError("Text formatter must be a str or a callable object.")
            self.model()._text_formatter[name] = fmt
        self.refreshTable()
        return None

    @setTextFormatter.server
    def setTextFormatter(self, name: str, fmt: Callable[[Any], str] | str):
        fmt = self.model()._text_formatter.get(name, None)
        return arguments(name, fmt)

    @_mgr.interface
    def setDataValidator(self, name: str, validator: Callable[[Any], None]):
        """Set a validator function to the column named `name`."""
        if validator is None:
            self.model()._validator.pop(name, None)
        else:
            if not callable(validator):
                raise TypeError("Validator must be callable.")
            self.model()._validator[name] = validator
        self.refreshTable()
        return None

    @setDataValidator.server
    def setDataValidator(self, name: str, validator: Callable[[Any], bool]):
        validator = self.model()._validator.get(name, None)
        return arguments(name, validator)

    def itemLabel(self, r: int, c: int):
        model = self.model()
        if content := model._decorations.get((r, c), None):
            return content[1]
        return None

    @_mgr.interface
    def setItemLabel(self, r: int, c: int, text: str | None):
        model = self.model()
        index = model.index(r, c)
        return model.set_cell_label(index, text)

    @setItemLabel.server
    def setItemLabel(self, r: int, c: int, text: str):
        return arguments(r, c, self.itemLabel(r, c))

    def setInCellSlot(self, pos: tuple[int, int], slot: InCellRangedSlot):
        """Set in-cell slot at the given position."""
        if slot is None:
            self._qtable_view._table_map.pop(pos)
        else:
            self._set_incell_slot(pos, slot)
        return None

    @_mgr.interface
    def _set_incell_slot(self, pos: tuple[int, int], slot: InCellRangedSlot):
        """Set in-cell slot at the given position undoably."""
        if slot is None:
            self._qtable_view._table_map.pop(pos, None)
        else:
            self._qtable_view._table_map[pos] = slot
            if dest := slot.last_destination:
                self._qtable_view._selection_model.set_ranges([dest])
        return None

    @_set_incell_slot.server
    def _set_incell_slot(self, pos: tuple[int, int], slot: InCellRangedSlot):
        slot = self._qtable_view._table_map.get(pos, None)
        return arguments(pos, slot)

    @_set_incell_slot.set_formatter
    def _set_incell_slot_fmt(self, pos, slot: InCellRangedSlot):
        return f"table.cell[{pos[0]}, {pos[1]}] = '&={slot.as_literal()}'"

    def refreshTable(self, process: bool = False) -> None:
        """Refresh table view."""
        self._qtable_view._update_all()
        if process:
            QtW.QApplication.processEvents()
        return None

    def undoStackView(self, show: bool = True):
        """Show undo stack viewer."""
        out = self._mgr.widget()
        if show:
            self.addSideWidget(out, name="Undo stack")
        return out

    def addSideWidget(self, widget: QtW.QWidget, name: str = "") -> None:
        """Add a widget to the side area of the table."""
        if self._side_area is None:
            from ._side_area import QTableSideArea

            area = QTableSideArea()
            self.addWidget(area)
            self._side_area = area
            self.setOrientation(self.orientation())

        self._side_area.addWidget(widget, name=name)
        self.setSizes([500, 200])
        return None

    def addOverlayWidget(
        self,
        widget: QtW.QWidget,
        label: str = "",
        topleft: tuple[float, float] = (0, 0),
        size: tuple[float, float] | None = None,
        grip: bool = True,
    ) -> None:
        """Add a widget as an overlay of the table."""
        from ._overlay import QOverlayFrame

        viewport = self._qtable_view.viewport()
        frame = QOverlayFrame(widget, viewport, grip=grip)
        frame.setLabel(label)
        frame.show()

        # if topleft is given as float, interpret as ratio in a cell
        top, left = topleft
        top_int, top_res = divmod(top, 1)
        left_int, left_res = divmod(left, 1)
        top_int = int(round(top_int))
        left_int = int(round(left_int))

        index_tl = self._qtable_view.model().index(top_int, left_int)

        rect_tl = self._qtable_view.visualRect(index_tl)
        pos = rect_tl.topLeft()
        pos.setX(pos.x() + rect_tl.width() * left_res)
        pos.setY(pos.y() + rect_tl.height() * top_res)
        frame.move(pos)

        table_size = self._qtable_view.size()
        if size is None:
            frame.resize(
                max(frame.width(), int(table_size.width() * 0.8)),
                max(frame.height(), int(table_size.height() * 0.8)),
            )
        else:
            width, height = size
            bottom_int, bottom_res = divmod(top + height - 1, 1)
            right_int, right_res = divmod(left + width - 1, 1)
            bottom_int = int(round(bottom_int))
            right_int = int(round(right_int))

            index_br = self._qtable_view.model().index(bottom_int, right_int)

            rect_br = self._qtable_view.visualRect(index_br)
            _w = rect_br.width() * right_res
            _h = rect_br.height() * bottom_res
            rect_br.setBottom(rect_br.bottom() + _h)
            rect_br.setRight(rect_br.right() + _w)
            rect = rect_tl.united(rect_br)
            frame.resize(rect.size())

        return frame

    def setDualView(self, orientation: str = "horizontal"):
        """Set dual view."""
        from ._table_wrappers import QTableDualView

        if orientation == "vertical":
            qori = Qt.Orientation.Vertical
        elif orientation == "horizontal":
            qori = Qt.Orientation.Horizontal
        else:
            raise ValueError("orientation must be 'vertical' or 'horizontal'.")

        widget0 = self.widget(0)
        if widget0 is not self._central_widget:
            widget0.setParent(None)
            widget0.deleteLater()
        self._central_widget.setParent(None)
        dual = QTableDualView(self._central_widget, qori)
        self.insertWidget(0, dual)
        self._qtable_view.setFocus()
        return dual

    def setPopupView(self):
        """Set splash view."""
        from ._table_wrappers import QTablePopupView

        widget0 = self.widget(0)
        if widget0 is not self._central_widget:
            widget0.setParent(None)
            widget0.deleteLater()
        self._central_widget.setParent(None)
        view = QTablePopupView(self._central_widget)
        self.insertWidget(0, view)
        view.exec()
        return view

    def resetViewMode(self) -> None:
        """Reset the view mode to the normal one."""
        widget0 = self.widget(0)
        if widget0 is not self._central_widget:
            widget0.setParent(None)
            self.insertWidget(0, self._central_widget)
            widget0.deleteLater()
        else:
            pass

        return None

    def moveToItem(
        self,
        row: int | None = None,
        column: int | None = None,
        clear_selection: bool = True,
    ) -> None:
        """Move current index."""
        selection_model = self._qtable_view._selection_model
        df_shape = self.model().df.shape
        table_shape = self.model().rowCount(), self.model().columnCount()

        if row is None:
            row = selection_model.current_index.row
        elif row < 0:
            if row == -1 and df_shape[0] - 1 <= selection_model.current_index.row:
                row += table_shape[0]
            else:
                row += df_shape[0]

        if column is None:
            column = selection_model.current_index.column
        elif column < 0:
            if column == -1 and df_shape[1] - 1 <= selection_model.current_index.column:
                column += table_shape[1]
            else:
                column += df_shape[1]

        # If selection is not going to be expanded, clear the selection first.
        if clear_selection or not (
            selection_model._ctrl_on or selection_model._shift_on
        ):
            selection_model.clear()

        selection_model.move_to(row, column)
        return None

    def showContextMenuAtIndex(self):
        """Programmatically show context menu at index (r, c)."""
        r, c = self._qtable_view._selection_model.current_index
        if r >= 0 and c >= 0:
            index = self._qtable_view.model().index(r, c)
            rect = self._qtable_view.visualRect(index)
            self.showContextMenu(rect.center())
        elif r < 0:
            header = self._qtable_view.horizontalHeader()
            rect = header.visualRectAtIndex(c)
            header._show_context_menu(rect.center())
        elif c < 0:
            header = self._qtable_view.verticalHeader()
            rect = header.visualRectAtIndex(r)
            header._show_context_menu(rect.center())
        else:
            raise RuntimeError("Invalid index")

    def raiseSlotError(self):
        qtable_view = self._qtable_view
        idx = qtable_view._selection_model.current_index
        if slot := qtable_view._table_map.get_by_dest(idx, None):
            if slot._current_error is not None:
                slot.raise_in_msgbox(parent=self)

    def tableStack(self) -> QTabbedTableStack | None:
        """Return the table stack."""
        try:
            stack = self.parentWidget().parentWidget()
        except AttributeError:
            stack = None
        if isinstance(stack, QtW.QTabWidget):
            # if a table is used in other widgets, it does not have a table stack
            # as a parent.
            return stack
        return None

    def parentViewer(self) -> _QtMainWidgetBase:
        """Return the parent table viewer."""
        if table_stack := self.tableStack():
            return table_stack.parentWidget()
        return None

    def screenshot(self):
        """Create an array of pixel data of the current view."""
        import qtpy

        img = self.grab().toImage()
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME.startswith("PySide"):
            arr = np.array(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr: np.ndarray = np.frombuffer(bits, np.uint8)
            arr = arr.reshape(h, w, c)

        return arr[:, :, [2, 1, 0, 3]]

    def _switch_head_and_index(self, axis: int = 1) -> None:
        """Switch the first row/column data and the index object."""
        self.setProxy(None)  # reset filter to avoid unexpected errors
        df = self.getDataFrame()
        if axis == 0:
            was_range = is_ranged(df.columns)
            if is_ranged(df.index):  # df[0] to index
                df_new = df.set_index(df.columns[0])
            else:  # index to df[0]
                df_new = df.reset_index()
            if was_range:
                _index = as_constructor(df.columns)(df_new.columns.size)
                df_new = df_new.set_axis(_index, axis=1)

        elif axis == 1:
            was_range = is_ranged(df.index)
            if is_ranged(df.columns):  # df[0] to column
                top_row = df.iloc[0, :].astype(str)
                df_new = df.iloc[1:, :]
                df_new = df_new.set_axis(top_row, axis=1)
            else:  # column to df[0]
                columns = range(len(df.columns))
                head = pd.DataFrame(
                    [
                        pd.Series([x], dtype=dtype)
                        for x, dtype in zip(df.columns, df.dtypes)
                    ],
                    columns=columns,
                )
                df = df.set_axis(columns, axis=1)
                df_new = pd.concat([head, df], axis=0)
            if was_range:
                df_new = df_new.set_axis(pd.RangeIndex(len(df_new)), axis=0)
        else:
            raise ValueError("axis must be 0 or 1.")
        self.setDataFrame(df_new)
        if axis == 0:
            self._qtable_view.resizeRowsToContents()
            if df.columns.size < df_new.columns.size:  # index to df[0]
                self._process_insert_columns(0, 1)
            elif df.columns.size > df_new.columns.size:  # df[0] to index
                self._process_remove_columns(0, 1)

        return None

    def _get_ref_expr(self, r: int, c: int) -> str | None:
        """Try to get a reference expression for the cell at (r, c)."""
        r = self._proxy.get_source_index(r)
        if slot := self._qtable_view._table_map.get((r, c), None):
            return slot.as_literal()
        return None

    def _get_ref_expr_by_dest(self, r: int, c: int) -> str | None:
        """Try to get a reference expression for the cell at (r, c)."""
        r = self._proxy.get_source_index(r)
        if slot := self._qtable_view._table_map.get_by_dest((r, c), None):
            return slot.as_literal()
        return None

    def _delete_selected_highlights(self) -> None:
        """Delete the selected highlight."""
        self._qtable_view._highlight_model.delete_selected()
        self._qtable_view._selection_model.set_ctrl(False)
        return None

    def _header_widgets(self) -> dict[int, QtW.QWidget]:
        """Return the header widgets."""
        return self._qtable_view.horizontalHeader()._header_widgets

    def _process_remove_rows(self, row: int, count: int):
        self._qtable_view._table_map.remove_rows(row, count)
        self._qtable_view._highlight_model.remove_rows(row, count)
        self._qtable_view._selection_model.remove_rows(row, count)

    def _process_insert_rows(self, row: int, count: int):
        self._qtable_view._table_map.insert_rows(row, count)
        self._qtable_view._selection_model.insert_rows(row, count)
        self._qtable_view._highlight_model.insert_rows(row, count)

    def _process_remove_columns(self, col: int, count: int):
        self._qtable_view._table_map.remove_columns(col, count)
        self._qtable_view._highlight_model.remove_columns(col, count)
        self._qtable_view._selection_model.remove_columns(col, count)

    def _process_insert_columns(self, col: int, count: int):
        self._qtable_view._table_map.insert_columns(col, count)
        self._qtable_view._selection_model.insert_columns(col, count)
        self._qtable_view._highlight_model.insert_columns(col, count)


class QMutableTable(QBaseTable):
    """A mutable table widget."""

    itemChangedSignal = Signal(ItemInfo)
    rowChangedSignal = Signal(HeaderInfo)
    columnChangedSignal = Signal(HeaderInfo)
    evaluatedSignal = Signal(EvalInfo)
    selectionChangedSignal = Signal()

    _data_raw: pd.DataFrame
    NaN = np.nan

    def __init__(
        self, parent: QtW.QWidget | None = None, data: pd.DataFrame | None = None
    ):
        super().__init__(parent, data)
        self._data_cache = None  # only used in SpreadSheet for now
        self.model().dataEdited.connect(self.setDataFrameValue)

        # header editing signals
        self._qtable_view.horizontalHeader().sectionDoubleClicked.connect(
            self.editHorizontalHeader
        )
        self._qtable_view.verticalHeader().sectionDoubleClicked.connect(
            self.editVerticalHeader
        )
        self._mgr.clear()

        @self.rowChangedSignal.connect
        def _on_row_changed(info: HeaderInfo):
            return self.setVerticalHeaderValue(info.index, info.value)

        @self.columnChangedSignal.connect
        def _on_col_changed(info: HeaderInfo):
            return self.setHorizontalHeaderValue(info.index, info.value)

    def tableShape(self) -> tuple[int, int]:
        """Return the available shape of the table."""
        model = self.model()
        nr = model.rowCount()
        nc = model.columnCount()
        return (nr, nc)

    def tableSlice(self) -> pd.DataFrame:
        """Return 2D table for display."""
        return self._data_raw

    def setDataFrameValue(self, r: _Sliceable, c: _Sliceable, value: Any) -> None:
        """Set `value` at array[r, c]."""
        if not self.isEditable():
            raise TableImmutableError("Table is immutable.")
        data = self._data_raw

        with self._mgr.merging(lambda _: self._set_value_fmt(r, c, value)):
            # convert values
            if isinstance(r, slice) and isinstance(c, slice):
                # delete references
                self._clear_incell_slots(r, c)
                _value = self._pre_set_array(r, c, value)
                _is_scalar = False
            else:
                self._set_incell_slot((r, c), None)
                _convert_value = self._get_converter(c)
                _value = _convert_value(c, value)
                _is_scalar = True

            # if table has proxy, indices must be adjusted
            r0 = self._get_proxy_source_index(r)

            _old_value = data.iloc[r0, c]
            if not _is_scalar:
                _old_value: pd.DataFrame
                _old_value = _old_value.copy()  # this is needed for undo

            if self._proxy.proxy_type != "none":
                # If table is filtered, the dataframe to be displayed is a different
                # object so we have to update it as well.
                self.model().updateValue(r, c, _value)

            # emit item changed signal if value changed
            if _was_changed(_value, _old_value) and self.isEditable():
                self._set_value(r0, c, r, c, value=_value, old_value=_old_value)

        return None

    def _clear_incell_slots(self, r: slice, c: slice) -> None:
        # this with-block is not needed but make it more efficient
        if len(self._qtable_view._table_map) < 128:
            for key in list(self._qtable_view._table_map.keys()):
                if r.start <= key[0] < r.stop and c.start <= key[1] < c.stop:
                    self._set_incell_slot(key, None)

        else:
            for _c in range(c.start, c.stop):
                for _r in range(r.start, r.stop):
                    self._set_incell_slot((_r, _c), None)
        return None

    def setLabeledData(self, r: slice, c: slice, value: pd.Series):
        """Set 1D data with index as labels."""

        if not self.isEditable():
            raise TableImmutableError("Table is immutable.")
        if c.stop != c.start + 1:
            raise ValueError("Only one column can be set at a time.")
        data = self._data_raw

        with self._mgr.merging(lambda _: self._set_value_fmt(r, c, value)):
            # delete references
            self._clear_incell_slots(r, c)

            _value = self._pre_set_array(r, c, pd.DataFrame(value))

            # if table has proxy, indices must be adjusted
            r0 = self._get_proxy_source_index(r)

            _old_value = data.iloc[r0, c].copy()  # this is needed for undo

            if self._proxy is not None:
                # If table is filtered, the dataframe to be displayed is a different
                # object so we have to update it as well.
                self.model().updateValue(r, c, _value)

            # emit item changed signal if value changed
            if _was_changed(_value, _old_value) and self.isEditable():
                self._set_value(r0, c, r, c, value=_value, old_value=_old_value)
                if isinstance(r0, slice):
                    _iter = range(r0.start, r0.stop)
                elif isinstance(r0, np.ndarray):
                    _iter = r0
                else:
                    raise UnreachableError(f"{type(r0)!r}.")
                for _i, _r in enumerate(_iter):
                    self.setItemLabel(_r, c.start, value.index[_i])
        return None

    def _get_proxy_source_index(self, r: int | slice):
        return self._proxy.get_source_index(r)

    def _pre_set_array(self, r: slice, c: slice, _value: pd.DataFrame):
        """Convert input dataframe for setting to data[r, c]."""
        if _value.size == 1:
            v = _value.values[0, 0]
            _value = self._data_raw.iloc[r, c].copy()
            for _ic, _c in enumerate(range(c.start, c.stop)):
                _convert_value = self._get_converter(_c)
                for _ir, _r in enumerate(range(r.start, r.stop)):
                    _value.iloc[_ir, _ic] = _convert_value(_c, v)
        else:
            for _ic, _c in enumerate(range(c.start, c.stop)):
                _convert_value = self._get_converter(_c)
                for _ir, _r in enumerate(range(r.start, r.stop)):
                    _value.iloc[_ir, _ic] = _convert_value(_c, _value.iloc[_ir, _ic])
        return _value

    @QBaseTable._mgr.undoable
    def _set_value(self, r, c, r_ori, c_ori, value, old_value):
        """Undoable set-value function."""
        self.updateValue(r, c, value)
        self._data_cache = None
        self._edited = True
        # update selection using the non-filtered index.
        self.setSelections([(r_ori, c_ori)])
        self.itemChangedSignal.emit(ItemInfo(r, c, value, old_value))
        return None

    @_set_value.undo_def
    def _set_value(self, r, c, r_ori, c_ori, value, old_value):
        self.updateValue(r, c, old_value)
        self._data_cache = None
        self.setSelections([(r_ori, c_ori)])
        self.itemChangedSignal.emit(ItemInfo(r, c, old_value, value))
        return None

    @_set_value.reduce_rule
    def _set_value_reduce(self, args_old, args_new):
        # r, c, r_ori, c_ori, value, old_value
        r = args_new["r"]
        c = args_new["c"]
        r_ori = args_new["r_ori"]
        c_ori = args_new["c_ori"]
        value = args_new["value"]
        old_value = args_old["old_value"]
        return arguments(r, c, r_ori, c_ori, value, old_value)

    def updateValue(self, r, c, value):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._data_raw.iloc[r, c] = value

        if self._proxy.proxy_type != "none":
            self._set_proxy(self._proxy)
        return self.refreshTable()

    def isEditable(self) -> bool:
        """Return the editability of the table."""
        return self.model()._editable

    @QBaseTable._mgr.interface
    def setEditable(self, editable: bool):
        """Set the editability of the table."""
        self.model()._editable = editable
        return None

    @setEditable.server
    def setEditable(self, editable: bool):
        return arguments(self.isEditable())

    @setEditable.set_formatter
    def _setEditable_fmt(self, editable: bool):
        return f"table.editable = {editable}"

    def toggleEditability(self) -> None:
        """Toggle editability of the table."""
        return self.setEditable(not self.isEditable())

    def connectItemChangedSignal(
        self,
        slot_val: Callable[[ItemInfo], None],
        slot_row: Callable[[HeaderInfo], None],
        slot_col: Callable[[HeaderInfo], None],
        slot_eval: Callable[[EvalInfo], None],
    ) -> None:
        self.itemChangedSignal.connect(slot_val)
        self.rowChangedSignal.connect(slot_row)
        self.columnChangedSignal.connect(slot_col)
        self.evaluatedSignal.connect(slot_eval)
        return None

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        keys = QtKeys(e)
        if not self._keymap.press_key(keys):
            return super().keyPressEvent(e)

    def _paste_data(self, df: pd.DataFrame):
        """Paste data to the table."""
        selections = self.selections()
        n_selections = len(selections)
        if n_selections == 0:
            return
        elif not self.isEditable():
            if stack := self.tableStack():
                stack.notifyEditability()
            return None
        elif n_selections > 1:
            raise SelectionRangeError("Cannot paste to multiple selections.")

        dr, dc = df.shape

        # check size and normalize selection slices
        sel = selections[0]
        rrange, crange = sel
        rstart = 0 if rrange.start is None else rrange.start
        rstop = rstart + dr if rrange.stop is None else rrange.stop
        cstart = 0 if crange.start is None else crange.start
        cstop = cstart + dc if crange.stop is None else crange.stop
        rrange = slice(rstart, rstop)
        crange = slice(cstart, cstop)
        sel = (rrange, crange)

        # check if the selection is valid
        rlen = rstop - rstart
        clen = cstop - cstart
        size = dr * dc

        if rlen * clen == 1 and size > 1:
            sel = (
                slice(rrange.start, rrange.start + dr),
                slice(crange.start, crange.start + dc),
            )

        elif size > 1 and (rlen, clen) != (dr, dc):
            # If selection is column-wide or row-wide, resize them
            model = self.model()
            if rlen == model.df.shape[0]:
                rrange = slice(0, dr)
                rlen = dr
            if clen == model.df.shape[1]:
                crange = slice(0, dc)
                clen = dc

            if (rlen, clen) != (dr, dc):
                raise SelectionRangeError(
                    f"Shape mismatch between data in clipboard {(rlen, clen)} and "
                    f"destination {(dr, dc)}."
                )
            else:
                sel = (rrange, crange)

        rsel, csel = sel

        # check dtype
        dtype_src = df.dtypes.values
        dtype_dst = self._data_raw.dtypes.values[csel]
        if any(a.kind != b.kind for a, b in zip(dtype_src, dtype_dst)):
            raise ValueError(
                f"Data type mismatch between data in clipboard {list(dtype_src)} and "
                f"destination {list(dtype_dst)}."
            )

        # update table
        self.setDataFrameValue(rsel, csel, df)
        self.setSelections([sel])

        return None

    def pasteFromClipBoard(self, sep=r"\s+"):
        """
        Paste data to table.

        This function supports many types of pasting.
        1. Single selection, single data in clipboard -> just paste
        2. Single selection, multiple data in clipboard -> paste starts from the
           selection position.
        3. Multiple selection, single data in clipboard -> paste the same value for
           all the selection.
        4. Multiple selection, multiple data in clipboard -> paste only if their
           shape is identical.

        Also, if data is filtrated, pasted data also follows the filtration.
        """
        df = self.readClipBoard(sep=sep)
        if df.shape == (1, 1) and QCellLiteralEdit._is_eval_like(str(df.iat[0, 0])):
            sels = self.selections()
            if len(sels) != 1:
                raise SelectionRangeError(
                    "Multiple ranges are selected. Failed to paste."
                )
            sel = sels[0]
            rsel, csel = sel
            if None in (rsel.start, rsel.stop, csel.start, csel.stop):
                raise SelectionRangeError("Selection is not valid. Failed to paste.")
            if rsel.stop - rsel.start == 1 and csel.stop - csel.start == 1:
                # single cell selection
                editor = self._qtable_view._edit_current()
                editor.setText(df.iat[0, 0])

        return self._paste_data(df)

    def deleteValues(self):
        """Replace selected cells with NaN."""
        if not self.isEditable():
            return None
        selections = self.selections()
        rsize, csize = self._data_raw.shape
        for sel in selections:
            # normalize selection range to non-None slices
            rsel, csel = sel
            rstart = 0 if rsel.start is None else rsel.start
            rstop = rsize if rsel.stop is None else rsel.stop
            cstart = 0 if csel.start is None else csel.start
            cstop = csize if csel.stop is None else csel.stop
            rsel = slice(rstart, rstop)
            csel = slice(cstart, cstop)

            nr = rstop - rstart
            nc = cstop - cstart
            dtypes = list(self._data_raw.dtypes.values[csel])
            df = pd.DataFrame(
                {
                    c: pd.Series(np.full(nr, self.NaN), dtype=dtypes[c])
                    for c in range(nc)
                },
            )
            self.setDataFrameValue(rsel, csel, df)
        return None

    def editHorizontalHeader(self, index: int) -> QHorizontalHeaderLineEdit:
        """Edit the horizontal header."""
        if not self.isEditable():
            if stack := self.tableStack():
                stack.notifyEditability()
            return None

        qtable = self._qtable_view
        _header = qtable.horizontalHeader()
        return QHorizontalHeaderLineEdit(parent=_header, table=self, pos=(-1, index))

    def editVerticalHeader(self, index: int) -> QVerticalHeaderLineEdit:
        """Edit the vertical header."""
        if not self.isEditable():
            if stack := self.tableStack():
                stack.notifyEditability()
            return None

        qtable = self._qtable_view
        _header = qtable.verticalHeader()
        return QVerticalHeaderLineEdit(parent=_header, table=self, pos=(index, -1))

    def setHorizontalHeaderValue(self, index: int, value: Any) -> None:
        return self._set_horizontal_header_value(index, value, constructor=None)

    @QBaseTable._mgr.interface
    def _set_horizontal_header_value(self, index: int, value, constructor=None):
        qtable = self._qtable_view
        _header = qtable.horizontalHeader()

        model = self.model()

        colname = model.df.columns[index]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model.rename_column(colname, value)

            _rename_column(model.df, index, value)
            self._filtered_columns = _rename_index(self._filtered_columns, index, value)
            if constructor is None:
                _rename_column(self._data_raw, index, value)
            else:
                self._data_raw.columns = constructor(self._data_raw.columns.size)

        self._edited = True

        # set selection
        self._qtable_view._selection_model.move_to(-1, index)

        # update
        self.refreshTable()

        # adjust header size
        size_hint = _header.sectionSizeHint(index)
        if _header.sectionSize(index) < size_hint:
            _header.resizeSection(index, size_hint)
        return None

    @_set_horizontal_header_value.server
    def _set_horizontal_header_value(self, index: int, value: Any, constructor) -> Any:
        return arguments(
            index,
            self.model().df.columns[index],
            constructor=as_constructor(self._data_raw.columns),
        )

    @_set_horizontal_header_value.set_formatter
    def _set_horizontal_header_value_fmt(self, index: int, value, constructor):
        return f"table.columns[{index}] = {value!r}"

    def setVerticalHeaderValue(self, index: int, value: Any) -> None:
        return self._set_vertical_header_value(index, value, constructor=None)

    @QBaseTable._mgr.interface
    def _set_vertical_header_value(self, index: int, value, constructor=None):
        qtable = self._qtable_view
        _header = qtable.verticalHeader()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _rename_row(self.model().df, index, value)
            r0 = self._get_proxy_source_index(index)
            self._filtered_index = _rename_index(self._filtered_index, r0, value)
            if constructor is None:
                _rename_row(self._data_raw, r0, value)
            else:
                self._data_raw.index = constructor(self._data_raw.index.size)

        self._edited = True

        # set selection
        self._qtable_view._selection_model.move_to(index, -1)

        # update
        self.refreshTable()

        # adjust size
        _width_hint = _header.sizeHint().width()
        _header.resize(QtCore.QSize(_width_hint, _header.height()))

        return None

    @_set_vertical_header_value.server
    def _set_vertical_header_value(self, index: int, value: Any, constructor):
        return arguments(
            index,
            self.model().df.index[index],
            constructor=as_constructor(self._data_raw.index),
        )

    @_set_vertical_header_value.set_formatter
    def _set_vertical_header_value_fmt(self, index: int, value, constructor):
        return f"table.index[{index}] = {value!r}"

    def undo(self) -> Any:
        """Undo last operation."""
        return self._mgr.undo()

    def redo(self) -> Any:
        """Redo last undo operation."""
        return self._mgr.redo()

    @staticmethod
    def _set_value_fmt(r, c, value):
        _r = fmt_slice(r)
        _c = fmt_slice(c)
        if isinstance(value, pd.DataFrame):
            if value.size < 6:
                _val = str(value.values.tolist())
            else:
                _val = "..."
        else:
            _val = fmt.map_object(value)
        return f"df.iloc[{_r}, {_c}] = {_val}"


class QMutableSimpleTable(QMutableTable):
    """A mutable table with a single QTableView."""

    def dataShapeRaw(self) -> tuple[int, int]:
        return self._data_raw.shape

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        return self._qtable_view_

    def createQTableView(self):
        from ._enhanced_table import _QTableViewEnhanced

        self._qtable_view_ = _QTableViewEnhanced(self)
        self.addWidget(self._qtable_view_)
        return None


def _was_changed(val: Any, old_val: Any) -> bool:
    # NOTE pd.NA == x returns pd.NA, not False
    out = False
    if isinstance(val, pd.DataFrame):
        out = True

    elif isna(val):
        if not isna(old_val):
            out = True
    else:
        if isna(old_val) or val != old_val:
            out = True
    return out


def _normalize_selections(
    selections: SelectionType, nr: int, nc: int
) -> list[tuple[slice, slice]]:
    _new_selections: list[tuple[slice, slice]] = []
    for sel in selections:
        r, c = sel
        # if int is used instead of slice
        r = _normalize_one(r, nr)
        c = _normalize_one(c, nc)
        _new_selections.append((r, c))
    return _new_selections


def _normalize_one(r, nr: int) -> slice:
    if not isinstance(r, slice):
        _r = r.__index__()
        if _r < 0:
            _r += nr
        r = slice(_r, _r + 1)
    else:
        start, stop, step = r.indices(nr)
        if step != 1:
            raise ValueError("step must be 1")
        r = slice(start, stop)
    return r


def _rename_row(df: pd.DataFrame, idx: int, new_name: str) -> None:
    """Rename row label at the given index."""
    rowname = df.index[idx]
    df.rename(index={rowname: new_name}, inplace=True)
    return None


def _rename_column(df: pd.DataFrame, idx: int, new_name: str) -> None:
    """Rename column label at the given index."""
    colname = df.columns[idx]
    df.rename(columns={colname: new_name}, inplace=True)
    as_not_ranged(df.columns)
    return None


def _rename_index(di: pd.Index | None, idx: int, new_name: str) -> pd.Index:
    if di is None:
        return di
    di_list = list(di)
    di_list[idx] = new_name
    return pd.Index(di_list)


def _fmt_slice(sl: slice) -> str:
    return f"{sl.start}:{sl.stop}"


def _selection_to_literal(sel: tuple[slice, slice]) -> str:
    rsel, csel = sel
    rsize = rsel.stop - rsel.start
    csize = csel.stop - csel.start
    if rsize == 1 and csize == 1:
        txt = f"[{rsel.start}, {csel.start}]"
    elif rsize == 1:
        txt = f"[{rsel.start}, {_fmt_slice(csel)}]"
    elif csize == 1:
        txt = f"[{_fmt_slice(rsel)}, {csel.start}]"
    else:
        txt = f"[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"
    return txt


_V = TypeVar("_V")


def _convert_value(
    c: int,
    x: str,
    validator: Callable[[_V], bool],
    converter: Callable[[int, int, str], _V],
) -> _V:
    """Convert value with validation."""
    val = converter(c, x)  # convert value first
    validator(val)  # Raise error if invalid
    return val


def _to_clipboard(s: str, excel: bool = True, **kwargs) -> None:
    """Copy dataframe to clipboard."""
    from pandas.io.clipboards import to_clipboard

    return to_clipboard(s, excel=excel, **kwargs)


def _tabulate(df: pd.DataFrame, tablefmt: str, index: bool = False) -> str:
    try:
        from tabulate import tabulate
    except ImportError:
        raise ImportError("No module named 'tabulate'. Please `pip install tabulate`.")

    return tabulate(
        df.astype("string").fillna(""),
        headers="keys",
        tablefmt=tablefmt,
        showindex=index,
    )


def _dataframe_to_grid_table(df: pd.DataFrame):
    cell_width = [col.str.len().max() for _, col in df.items()]

    def table_div(sep="-"):
        return "".join("+" + (width + 2) * sep for width in cell_width) + "+"

    def row_to_str(row: Iterable[str]):
        return "| " + " | ".join([f"{x:{w}}" for x, w in zip(row, cell_width)]) + " |"

    lines = [table_div(), row_to_str(df.columns), table_div("=")]
    for _, row in df.iterrows():
        lines.append(row_to_str(row))
        lines.append(table_div())
    return "\n".join(lines)



from __future__ import annotations
from abc import abstractmethod
from contextlib import contextmanager
from typing import (
    Hashable,
    TYPE_CHECKING,
    MutableSequence,
    Sequence,
    TypeVar,
    Tuple,
    Iterator,
)

from tabulous.types import _SingleSelection, SelectionType
from ._base import TableComponent

if TYPE_CHECKING:
    import pandas as pd


T = TypeVar("T")
_Range = Tuple[slice, slice]


class _TableRanges(TableComponent, MutableSequence[_Range]):
    def __init__(self, parent: T = TableComponent._no_ref):
        super().__init__(parent)
        self._is_blocked = False

    @abstractmethod
    def _get_list(self) -> list[_Range]:
        """Get the list of ranges."""

    @abstractmethod
    def update(self, val: list[_Range]) -> None:
        """Set a list of ranges."""

    def __repr__(self) -> str:
        rng_str: list[str] = []
        for rng in self:
            r, c = rng
            rng_str.append(f"[{_fmt_slice(r)}, {_fmt_slice(c)}]")
        return f"{self.__class__.__name__}({', '.join(rng_str)})"

    def __getitem__(self, index: int) -> _Range:
        """The selected range at the given index."""
        return self._get_list()[index]

    def __setitem__(self, index: int, val: _SingleSelection) -> None:
        if self._is_blocked:
            raise RuntimeError("Cannot set item while blocked.")
        lst = self._get_list()
        lst[index] = val
        return self.update(lst)

    def __delitem__(self, index: int) -> None:
        if self._is_blocked:
            raise RuntimeError("Cannot delete item while blocked.")
        lst = self._get_list()
        del lst[index]
        return self.update(lst)

    def __len__(self) -> int:
        """Number of selections."""
        return len(self._get_list())

    def __iter__(self):
        """Iterate over the selection ranges."""
        return iter(self._get_list())

    def _set_value(self, value: SelectionType):
        if self._is_blocked:
            raise RuntimeError("Cannot set item(s) while blocked.")
        if not isinstance(value, list):
            value = [value]
        return self.update(value)

    @property
    def values(self) -> SelectedData:
        return SelectedData(self)

    def insert(self, index: int, rng: _Range) -> None:
        if self._is_blocked:
            raise RuntimeError("Cannot insert item while blocked.")
        lst = self._get_list()
        lst.insert(index, rng)
        return self.update(lst)

    @contextmanager
    def blocked(self, block: bool = True):
        _was_blocked = self._is_blocked
        self._is_blocked = block
        try:
            yield
        finally:
            self._is_blocked = _was_blocked

    def block(self, block: bool = True) -> bool:
        """Block or unblock changing ranges."""
        self._is_blocked = block
        return block


class SelectionRanges(_TableRanges):
    """A table data specific selection range list."""

    def _get_list(self):
        return list(self.parent._qwidget.selections())

    def update(self, value: SelectionRanges):
        """Update the selection ranges."""
        return self.parent._qwidget.setSelections(value)


class HighlightRanges(_TableRanges):
    """A table data specific highlight list."""

    def _get_list(self):
        return list(self.parent._qwidget.highlights())

    def update(self, value: HighlightRanges):
        """Update the highlight ranges."""
        return self.parent._qwidget.setHighlights(value)


class SelectedData(Sequence["pd.DataFrame"]):
    """Interface with the selected data."""

    def __init__(self, obj: SelectionRanges):
        self._obj = obj

    def __getitem__(self, index: int) -> pd.DataFrame:
        """Get the selected data at the given index of selection."""
        data = self._obj.parent.data_shown
        sl = self._obj[index]
        return data.iloc[sl]

    def __len__(self) -> int:
        """Number of selections."""
        return len(self._obj)

    def __iter__(self) -> Iterator[pd.DataFrame]:
        return (self[i] for i in range(len(self)))

    def itercolumns(self) -> Iterator[tuple[Hashable, pd.Series]]:
        all_data: dict[Hashable, pd.Series] = {}
        import pandas as pd

        for data in self:
            for col in data.columns:
                if col in all_data.keys():
                    all_data[col] = pd.concat([all_data[col], data[col]])
                else:
                    all_data[col] = data[col]
        return iter(all_data.items())


def _fmt_slice(sl: slice) -> str:
    s0 = sl.start if sl.start is not None else ""
    s1 = sl.stop if sl.stop is not None else ""
    return f"{s0}:{s1}"



from __future__ import annotations
from typing import Any
from collections_undo import UndoManager, fmt
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt
from qtpy.sip import isdeleted

import numpy as np
import pandas as pd

from ._qt_const import MonospaceFontFamily

MonospaceFont = QtGui.QFont(MonospaceFontFamily)


class QUndoStackViewer(QtW.QWidget):
    """A viewer of the undo stack"""

    def __init__(self, mgr: UndoManager):
        super().__init__()
        self._mgr = mgr
        self._listview = QtW.QListView()
        self._listview.setModel(QUndoStackModel(mgr))
        mgr.called.append(self.refresh)

        self._undo_button = QtW.QPushButton("Undo")
        self._redo_button = QtW.QPushButton("Redo")
        self._undo_button.setToolTip("Undo the last action")
        self._redo_button.setToolTip("Redo the last undone action")

        _layout = QtW.QVBoxLayout()
        _layout.addWidget(self._listview)
        self.setLayout(_layout)

        _footer = QtW.QWidget()
        _footer_layout = QtW.QHBoxLayout()
        _footer_layout.setContentsMargins(0, 0, 0, 0)
        _footer_layout.addWidget(self._undo_button)
        _footer_layout.addWidget(self._redo_button)
        _footer.setLayout(_footer_layout)

        _layout.addWidget(_footer)

        self._undo_button.clicked.connect(self._mgr.undo)
        self._redo_button.clicked.connect(self._mgr.redo)
        self.setMinimumHeight(160)

    def refresh(self, *_):
        if isdeleted(self):
            self._mgr.called.remove(self.refresh)
        else:
            self._listview.model().updateUndoShape()
            self._listview.viewport().update()


class QUndoStackModel(QtCore.QAbstractListModel):
    def __init__(self, mgr: UndoManager) -> None:
        super().__init__()
        self._mgr = mgr
        self._current_size = 0

    def data(
        self, index: QtCore.QModelIndex, role: int = Qt.ItemDataRole.DisplayRole
    ) -> Any:
        if not index.isValid():
            return QtCore.QVariant()
        if role == Qt.ItemDataRole.DisplayRole:
            r = index.row()
            stack = self._mgr.stack_undo
            if r < len(stack):
                try:
                    desc = stack[r].format()
                except Exception as e:
                    desc = f"{type(e).__name__}: {e}"
                return desc

        elif role == Qt.ItemDataRole.FontRole:
            return MonospaceFont

        return QtCore.QVariant()

    def updateUndoShape(self):
        nr = self._mgr.stack_lengths[0]
        nr0 = self._current_size
        if nr0 < nr:
            self.beginInsertRows(QtCore.QModelIndex(), nr0, nr - 1)
            self.insertRows(nr0, nr - nr0, QtCore.QModelIndex())
            self.endInsertRows()
        elif nr < nr0:
            self.beginRemoveRows(QtCore.QModelIndex(), nr, nr0 - 1)
            self.removeRows(nr, nr0 - nr, QtCore.QModelIndex())
            self.endRemoveRows()
        self._current_size = nr

    def rowCount(self, parent: QtCore.QModelIndex = QtCore.QModelIndex()) -> int:
        return len(self._mgr.stack_undo)

    def flags(self, index):
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable


class QtUndoManager(UndoManager):
    def __init__(self, maxsize=1e7) -> None:
        super().__init__(measure=_count_data_size, maxsize=maxsize)
        self._widget = None

    def widget(self) -> QUndoStackViewer:
        if self._widget is None or isdeleted(self._widget):
            self._widget = QUndoStackViewer(self)
        return self._widget


def _count_data_size(*args, **kwargs) -> float:
    total_nbytes = 0
    for arg in args:
        total_nbytes += _getsizeof(arg)
    for v in kwargs.values():
        total_nbytes += _getsizeof(v)
    return total_nbytes


def _getsizeof(obj) -> float:
    if isinstance(obj, pd.DataFrame):
        if obj.size == 0:
            # The dtype might be object
            nbytes = 0
        else:
            nbytes = obj.memory_usage(deep=True).sum()
    elif isinstance(obj, pd.Series):
        nbytes = obj.memory_usage(deep=True)
    elif isinstance(obj, np.ndarray):
        nbytes = obj.nbytes
    elif isinstance(obj, (list, tuple, set)):
        nbytes = sum(_getsizeof(x) for x in obj)
    elif isinstance(obj, dict):
        nbytes = sum(_getsizeof(x) for x in obj.values())
    else:
        nbytes = 1  # approximate
    return nbytes


def fmt_slice(sl: slice) -> str:
    if not isinstance(sl, slice):
        return fmt.map_object(sl)
    return f"{sl.start}:{sl.stop}"



from __future__ import annotations
from typing import Sequence, TYPE_CHECKING

from qtpy import QtWidgets as QtW
from qtpy.QtCore import Signal

from ._callback import BoundCallback
from ._keymap_objects import QtKeys, QtKeyMap


class QtKeyMapView(QtW.QWidget):
    """A viewer widget for keymap."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Keymaps")
        _layout = QtW.QVBoxLayout(self)
        self._list = QKeyMapList()
        self._keyseq_edit = QKeyComboEdit()
        _layout.addWidget(self._list)
        _layout.addWidget(self._keyseq_edit)
        self._keyseq_edit.seqChanged.connect(self._list.filter)

        self.setLayout(_layout)
        self._layout = _layout
        self.setMinimumWidth(520)

    @classmethod
    def from_keymap(cls, kmap: QtKeyMap) -> QtKeyMapView:
        self = cls()
        self.loadKeyMap(kmap)
        return self

    def loadKeyMap(self, kmap: QtKeyMap, offset: tuple[QtKeys, ...] = ()):
        for key, child in kmap.items():
            current_keys = offset + (key,)
            if isinstance(child, QtKeyMap):
                if (c := child.activated_callback) is not None:
                    self._list.addKeyMapItem(current_keys, c)
                self.loadKeyMap(child, current_keys)
            else:
                self._list.addKeyMapItem(list(current_keys), child)
        return None


class QKeySequenceEdit(QtW.QKeySequenceEdit):
    def timerEvent(self, a0) -> None:
        return None

    def keyCombo(self) -> list[QtKeys]:
        seq = self.keySequence()
        return [QtKeys(s) for s in seq.toString().split(", ") if s]


class QKeyComboEdit(QtW.QWidget):
    seqChanged = Signal(list)

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        _layout = QtW.QHBoxLayout(self)
        _layout.setContentsMargins(0, 0, 0, 0)

        label = QtW.QLabel("Filter: ")
        seq = QKeySequenceEdit()
        apply_btn = QtW.QPushButton("Apply")
        clear_btn = QtW.QPushButton("Clear")

        _layout.addWidget(label)
        _layout.addWidget(seq)
        _layout.addWidget(apply_btn)
        _layout.addWidget(clear_btn)

        self.setLayout(_layout)

        apply_btn.clicked.connect(self.emitKeySequence)
        clear_btn.clicked.connect(self.clear)

        self._keyseq_edit = seq

    def emitKeySequence(self):
        return self.seqChanged.emit(self._keyseq_edit.keyCombo())

    def clear(self):
        self._keyseq_edit.clear()
        self.emitKeySequence()


# ##############################################################################
#    List widget of all the keycombos
# ##############################################################################


class QKeyMapList(QtW.QListWidget):
    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.NoSelection)

    def addKeyMapItem(self, keys: list[QtKeys], callback: BoundCallback):
        item = QtW.QListWidgetItem()
        self.addItem(item)
        self.setItemWidget(item, QtKeyBindItem(key=keys, desc=callback.desc))
        return None

    def filter(self, keys: QtKeys | list[QtKeys]):
        if isinstance(keys, QtKeys):
            keys = [keys]
        for i in range(self.count()):
            item = self.item(i)
            widget = self.itemWidget(item)
            item.setHidden(not widget.startswith(keys))
        return None

    if TYPE_CHECKING:

        def itemWidget(self, item: QtW.QListWidgetItem) -> QtKeyBindItem:
            ...


class QtKeyBindItem(QtW.QWidget):
    """Item widget for QKeyMapList."""

    def __init__(
        self, parent=None, key: QtKeys | Sequence[QtKeys] | None = None, desc: str = ""
    ):
        super().__init__(parent)
        _layout = QtW.QHBoxLayout()
        _layout.setContentsMargins(2, 0, 2, 0)
        self.setLayout(_layout)
        self._key_label = QtW.QLabel()
        self._desc = QtW.QLabel()
        _layout.addWidget(self._key_label)
        _layout.addWidget(self._desc)

        self.setKeyText(key)
        self.setDescription(desc)
        self.key = key

    def setKeyText(self, key: QtKeys | Sequence[QtKeys]) -> None:
        if isinstance(key, QtKeys):
            text = f"<code>{key}</code>"
        else:
            text = " &rarr; ".join(f"<code>{k}</code>" for k in key)
        self._key_label.setText(text)
        return None

    def setDescription(self, desc: str) -> None:
        return self._desc.setText(desc)

    def startswith(self, prefix: list[QtKeys]) -> bool:
        """True if the key sequence starts with the given prefix."""
        nkey_pref = len(prefix)
        if nkey_pref == 0:
            return True
        elif isinstance(self.key, QtKeys):
            return nkey_pref == 1 and self.key == prefix[0]
        else:
            nkey = len(self.key)
            return nkey_pref <= nkey and prefix == self.key[:nkey_pref]



from __future__ import annotations
import re
from typing import (
    Any,
    Callable,
    Hashable,
    Iterator,
    Literal,
    MutableMapping,
    Sequence,
    TYPE_CHECKING,
    TypeVar,
    Union,
    overload,
)

from qtpy import QtGui
from qtpy.QtCore import Qt
from functools import reduce
from operator import or_

from tabulous.exceptions import TriggerParent

from ._callback import BoundCallback

if TYPE_CHECKING:
    from typing_extensions import Self


NoModifier = Qt.KeyboardModifier.NoModifier
Ctrl = Qt.KeyboardModifier.ControlModifier
Shift = Qt.KeyboardModifier.ShiftModifier
Alt = Qt.KeyboardModifier.AltModifier
Meta = Qt.KeyboardModifier.MetaModifier


_MODIFIERS = {
    "None": NoModifier,
    "Shift": Shift,
    "Ctrl": Ctrl,
    "Control": Ctrl,
    "Alt": Alt,
    "Meta": Meta,
}

_SYMBOLS = {
    "Esc": "Escape",
    "!": "Exclam",
    '"': "QuoteDbl",
    "#": "NumberSign",
    "$": "Dollar",
    "%": "Percent",
    "&": "Ampersand",
    "'": "Apostrophe",
    "(": "ParenLeft",
    ")": "ParenRight",
    "*": "Asterisk",
    "+": "Plus",
    ",": "Comma",
    "-": "Minus",
    ".": "Period",
    "/": "Slash",
    ":": "Colon",
    ";": "Semicolon",
    "<": "Less",
    "=": "Equal",
    ">": "Greater",
    "?": "Question",
    "@": "At",
    "[": "BracketLeft",
    "\\": "Backslash",
    "]": "BracketRight",
    "^": "AsciiCircum",
    "_": "Underscore",
    "`": "QuoteLeft",
    "{": "BraceLeft",
    "|": "Bar",
    "}": "BraceRight",
    "~": "AsciiTilde",
}

_MODIFIERS_INV = {
    NoModifier: "None",
    Shift: "Shift",
    Ctrl: "Ctrl",
    Alt: "Alt",
    Meta: "Meta",
    Ctrl | Shift: "Ctrl+Shift",
    Ctrl | Alt: "Ctrl+Alt",
    Alt | Shift: "Alt+Shift",
}


def _parse_string(keys: str):
    if keys in _MODIFIERS_INV.values():
        mods = keys.split("+")
        qtmod = reduce(or_, [_MODIFIERS[m] for m in mods])
        return qtmod, ExtKey.No

    *mods, btn = keys.split("+")

    # get modifiler
    if not mods:
        qtmod = Qt.KeyboardModifier.NoModifier
    else:
        qtmod = reduce(or_, [_MODIFIERS[m] for m in mods])

    # get button
    if btn in _SYMBOLS:
        btn = _SYMBOLS[btn]

    if btn != "{}":
        qtkey = getattr(Qt.Key, f"Key_{btn}")
    else:
        qtkey = ExtKey.Any
    return qtmod, qtkey


def _parse_modifiers(km: Qt.KeyboardModifiers) -> Qt.KeyboardModifiers:
    """Keypad modifier should be ignored."""
    if km & Qt.KeyboardModifier.KeypadModifier:
        km ^= Qt.KeyboardModifier.KeypadModifier

    return km


class ExtKey:
    """Key extension."""

    No = -1
    Any = -2


KeyType = Union[QtGui.QKeyEvent, "QtKeys", str]
MODIFIER_KEYS = frozenset(
    {Qt.Key.Key_Shift, Qt.Key.Key_Control, Qt.Key.Key_Meta, Qt.Key.Key_Alt}
)
ARROW_KEYS = frozenset(
    {Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_Left, Qt.Key.Key_Right}
)
FUNC_ARROW_KEYS = frozenset(
    {Qt.Key.Key_Home, Qt.Key.Key_End, Qt.Key.Key_PageUp, Qt.Key.Key_PageDown}
)
ALPHA = QtGui.QKeySequence("α")[0]
OMEGA = QtGui.QKeySequence("ω")[0]

CYR_A = QtGui.QKeySequence("а")[0]
CYR_YA = QtGui.QKeySequence("я")[0]


class QtKeys:
    """A custom class for handling key events."""

    def __init__(self, e: KeyType):
        if isinstance(e, QtGui.QKeyEvent):
            self.modifier = _parse_modifiers(e.modifiers())
            self.key = e.key()
            if self.key in MODIFIER_KEYS:  # modifier only
                self.key = ExtKey.No
        elif isinstance(e, str):
            mod, key = _parse_string(e)
            self.modifier = mod
            self.key = key
        elif isinstance(e, QtKeys):
            self.modifier = e.modifier
            self.key = e.key
        else:
            raise TypeError(
                "QtKeys can only be initialized with QKeyEvent, str or QtKeys, "
                f"but got {type(e)}."
            )

    def __hash__(self) -> int:
        return hash((self.modifier, self.key))

    def _reduce_key(self) -> Self:
        if self.key == ExtKey.No:
            # this case is needed to avoid triggering parametric key binding
            # with modifiers.
            return self
        new = QtKeys(self)
        new.key = ExtKey.Any
        return new

    def __str__(self) -> str:
        mod = _MODIFIERS_INV.get(self.modifier, "???")
        keystr = self.key_string()
        if mod == "None":
            return keystr
        elif keystr == "":
            return mod
        return "+".join([mod, keystr])

    def __repr__(self):
        return f"QtKeys({str(self)})"

    def __eq__(self, other):
        if isinstance(other, QtKeys):
            return self.modifier == other.modifier and self.key == other.key
        elif isinstance(other, str):
            qtmod, qtkey = _parse_string(other)
            return self.modifier == qtmod and self.key == qtkey
        else:
            raise TypeError("QtKeys can only be compared to QtKeys or str")

    def is_typing(self) -> bool:
        """True if key is a letter or number."""
        _modifier_ok = self.modifier in (
            Qt.KeyboardModifier.NoModifier,
            Qt.KeyboardModifier.ShiftModifier,
        )
        _key_ok = (
            Qt.Key.Key_Exclam <= self.key <= Qt.Key.Key_ydiaeresis
            or ALPHA <= self.key <= OMEGA
            or CYR_A <= self.key <= CYR_YA
        )
        return _modifier_ok and _key_ok

    def is_moving(self) -> bool:
        """True if arrows are pushed."""
        return self.key in ARROW_KEYS

    def is_moving_func(self) -> bool:
        """True if function arrows are pushed."""
        return self.key in FUNC_ARROW_KEYS

    def key_string(self, check_shift: bool = False) -> str:
        """Get clicked key in string form."""
        if self.key == -1:
            return ""
        elif self.key == -2:
            return "{}"

        s = QtGui.QKeySequence(self.key).toString()
        if check_shift and not self.has_shift():
            s = s.lower()
        return s

    def has_ctrl(self) -> bool:
        """True if Ctrl is pressed."""
        return bool(self.modifier & Qt.KeyboardModifier.ControlModifier)

    def has_shift(self) -> bool:
        """True if Shift is pressed."""
        return bool(self.modifier & Qt.KeyboardModifier.ShiftModifier)

    def has_alt(self) -> bool:
        """True if Alt is pressed."""
        return bool(self.modifier & Qt.KeyboardModifier.AltModifier)

    def has_key(self) -> bool:
        """True if non-modifier key is pressed."""
        return self.key != ExtKey.No


_K = TypeVar("_K", bound=Hashable)
_V = TypeVar("_V")


class RecursiveMapping(MutableMapping[_K, _V]):
    if TYPE_CHECKING:
        from abc import abstractmethod

        @abstractmethod
        def __getitem__(self, key: _K) -> _V | Self[_K, _V]:
            ...

        def get(self, key: _K, default: _V) -> _V | Self[_K, _V]:
            ...

        @abstractmethod
        def __iter__(self) -> Iterator[_V | Self[_K, _V]]:
            ...

        @abstractmethod
        def items(self) -> Iterator[tuple[_K, _V | Self[_K, _V]]]:
            ...


_F = TypeVar("_F", bound=Callable)


def _normalize_key_combo(s: list[str] | str) -> list[str] | str:
    if not isinstance(s, str):
        return s
    s = s.replace(" ", "")  # delete space
    if re.match(".+(,.+)+", s):
        # "Ctrl+A,Ctrl+C" matches, "Ctrl+," doesn't
        return s.split(",")
    return s


class QtKeyMap(RecursiveMapping[QtKeys, Callable]):
    """A mapping object for key map and key combo registration."""

    def __init__(
        self,
        instance: Any | None = None,
        activated: Callable | None = None,
        deactivated: Callable | None = None,
    ):
        self._keymap: dict[QtKeys, QtKeyMap | Callable] = {}
        self._current_map: QtKeyMap | None = None
        self._current_combo: list[QtKeys] = []
        self._activated_callback = activated or _DUMMY_CALLBACK
        self._deactivated_callback = deactivated or _DUMMY_CALLBACK
        self._released_callback = _DUMMY_CALLBACK
        self._instances: dict[int, QtKeyMap] = {}
        self._obj = instance

    @property
    def current_map(self) -> Self:
        """Return the current active keymap object."""
        if self._current_map is None:
            return self
        return self._current_map

    @property
    def previous_map(self) -> Self | None:
        """Return the parent of the current active keymap object."""
        if len(self._current_combo) == 0:
            return None
        current = self
        for k in self._current_combo[:-1]:
            current = current._keymap[k]
        return current

    @property
    def activated_callback(self) -> BoundCallback | None:
        """The function to be called when the keymap is activated."""
        if (a := self._activated_callback) is not _DUMMY_CALLBACK:
            return a
        return None

    @property
    def deactivated_callback(self) -> BoundCallback | None:
        """The function to be called when the keymap is deactivated."""
        if (a := self._deactivated_callback) is not _DUMMY_CALLBACK:
            return a
        return None

    @property
    def last_pressed(self) -> QtKeys | None:
        """The last pressed key."""
        if self._current_combo:
            return self._current_combo[-1]
        return None

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        _id = id(obj)
        if (new := self._instances.get(_id, None)) is None:
            new = self.copy()
            new._obj = obj
            self._instances[_id] = new
        return new

    def copy(self) -> Self:
        """Return a copy of an instance."""
        new = self.__class__(
            self._obj, self._activated_callback, self._deactivated_callback
        )
        new._keymap = self._keymap.copy()
        return new

    def __getitem__(self, key: KeyType):
        return self._keymap[QtKeys(key)]

    def _repr(self, indent: int = 0) -> str:
        indent_str = " " * indent
        strings = []
        if self._activated_callback is not _DUMMY_CALLBACK:
            strings.append(f"<activated>: {self._activated_callback!r}")
        if self._deactivated_callback is not _DUMMY_CALLBACK:
            strings.append(f"<deactivated>: {self._deactivated_callback!r}")
        for k, v in self.items():
            if isinstance(v, QtKeyMap):
                vrepr = v._repr(indent + 2)
            else:
                vrepr = repr(v)
            strings.append(f"{k}: {vrepr}")
        strings = f",\n{indent_str}  ".join(strings)
        return f"QtKeyMap(\n{indent_str}  {strings}\n{indent_str})"

    def __repr__(self) -> str:
        return self._repr()

    def add_child(self, key: KeyType, child_callback: Callable | None = None) -> None:
        """Add a child keymap."""
        kmap = QtKeyMap()
        if child_callback is not None:
            kmap._activated_callback = BoundCallback(child_callback, keys=key)
        self[key] = kmap
        return None

    @overload
    def bind(
        self,
        key: KeyType | Sequence[KeyType],
        callback: Literal[None] = None,
        overwrite: bool = False,
        desc: str | None = None,
        **kwargs,
    ) -> Callable[[_F], _F]:
        ...

    @overload
    def bind(
        self,
        key: KeyType | Sequence[KeyType],
        callback: _F,
        overwrite: bool = False,
        desc: str | None = None,
        **kwargs,
    ) -> _F:
        ...

    def bind(self, key, callback=None, *, overwrite=False, desc=None, **kwargs):
        """Assign key or key combo to callback."""

        def wrapper(func):
            _key = _normalize_key_combo(key)
            _func = BoundCallback(func, keys=_key, desc=desc, kwargs=kwargs)

            if isinstance(_key, (str, QtKeys)):
                self._bind_key(_key, _func, overwrite)
            elif isinstance(_key, Sequence):
                self._bind_keycombo(_key, _func, overwrite)
            else:
                raise TypeError("key must be a string or a sequence of strings")
            return func

        return wrapper if callback is None else wrapper(callback)

    def unbind(self, key: str) -> None:
        _key = _normalize_key_combo(key)
        if isinstance(_key, (str, QtKeys)):
            del self.current_map[_key]
        elif isinstance(_key, Sequence):
            current = self
            *pref, last = _key
            for k in pref:
                current = current[k]
            del current[last]
        else:
            raise TypeError("key must be a string or a sequence of strings")
        return None

    @overload
    def bind_deactivated(
        self,
        key: KeyType | Sequence[KeyType],
        callback: Literal[None] = None,
        overwrite: bool = False,
        **kwargs,
    ) -> Callable[[_F], _F]:
        ...

    @overload
    def bind_deactivated(
        self,
        key: KeyType | Sequence[KeyType],
        callback: _F,
        overwrite: bool = False,
        **kwargs,
    ) -> _F:
        ...

    def bind_deactivated(self, key, callback=None, overwrite=False, **kwargs):
        def wrapper(func):
            _key = _normalize_key_combo(key)
            _func = BoundCallback(func, keys=_key, kwargs=kwargs)

            if isinstance(_key, (str, QtKeys)):
                _key = [_key]
            current = self
            for i, k in enumerate(_key):
                if not isinstance(current, QtKeyMap):
                    seq = _key[:i]
                    raise ValueError(
                        f"Non keymap object {type(current)} encountered at "
                        f"{', '.join(seq)}."
                    )
                k = QtKeys(k)
                try:
                    current = current[k]
                except KeyError:
                    raise KeyError(f"Key {k} not found in {current!r}")
            if current._deactivated_callback is not _DUMMY_CALLBACK and not overwrite:
                raise ValueError(f"Key {key} already exists")
            current._deactivated_callback = _func
            return func

        return wrapper if callback is None else wrapper(callback)

    def bind_released(self, key, callback=None, overwrite=False, desc=None, **kwargs):
        def wrapper(func):
            _key = _normalize_key_combo(key)
            _func = BoundCallback(func, keys=_key, kwargs=kwargs)

            if isinstance(_key, (str, QtKeys)):
                _key = [_key]
            current = self
            for i, k in enumerate(_key):
                if not isinstance(current, QtKeyMap):
                    seq = _key[:i]
                    raise ValueError(
                        f"Non keymap object {type(current)} encountered "
                        f"at {', '.join(seq)}."
                    )
                k = QtKeys(k)
                try:
                    current = current[k]
                except KeyError:
                    raise KeyError(f"Key {k} not found in {current!r}")
            if current._released_callback is not _DUMMY_CALLBACK and not overwrite:
                raise ValueError(f"Key {key} already exists")
            current._released_callback = _func
            return func

        return wrapper if callback is None else wrapper(callback)

    def rebind(
        self,
        old: KeyType | Sequence[KeyType],
        new: KeyType | Sequence[KeyType],
        overwrite: bool = False,
    ):
        old = _normalize_key_combo(old)
        new = _normalize_key_combo(new)
        if isinstance(old, (str, QtKeys, QtGui.QKeyEvent)):
            old = [old]
        if isinstance(new, (str, QtKeys, QtGui.QKeyEvent)):
            new = [new]
        src = self
        for k in old[:-1]:
            src = src[QtKeys(k)]
        self.bind(new, src.pop(old[-1]), overwrite=overwrite)
        return None

    def _bind_key(
        self, key: KeyType, callback: BoundCallback, overwrite: bool = False
    ) -> None:
        child = self.get(key, None)
        if callable(child):
            if not overwrite:
                raise ValueError(f"A callback is already assigned to key {key}.")
            self[key] = callback

        elif isinstance(child, QtKeyMap):
            if not overwrite and child._activated_callback is not _DUMMY_CALLBACK:
                raise ValueError(
                    f"A keymap with callback is already assigned to key {key}."
                )
            child._activated_callback = callback

        else:
            self[key] = callback

        return None

    def _bind_keycombo(
        self, combo: Sequence[KeyType], callback: BoundCallback, overwrite: bool = False
    ) -> None:
        if isinstance(combo, str):
            raise TypeError("Combo must be an iterable of keys")
        current = self
        *pref, last = combo
        for key in pref:
            key = QtKeys(key)
            if key not in current:
                current.add_child(key)

            child = current[key]
            if callable(child):
                child = self.__class__(self._obj, activated=child)
                current[key] = child
            current = child

        current._bind_key(last, callback, overwrite)
        return None

    def press_key(self, key: KeyType) -> bool:
        """
        Emulate pressing a key.

        Parameters
        ----------
        key : KeyType
            Key expression to press.

        Returns
        -------
        bool
            True if the key triggered anything, False otherwise.
        """
        key = _normalize_key_combo(key)
        if isinstance(key, (str, QtKeys, QtGui.QKeyEvent)):
            return self._press_one_key(QtKeys(key))
        else:
            for k in key:
                out = self._press_one_key(QtKeys(k))
            return out

    def release_key(self) -> None:
        current = self.current_map
        if current._obj is None:
            current._released_callback()
        else:
            current._released_callback(current._obj)
        return None

    def _press_one_key(self, key: QtKeys) -> bool:
        current = self.current_map
        _is_parametric = False
        try:
            callback = current[key]
        except KeyError:
            try:
                callback = current[key._reduce_key()]
            except KeyError:
                # Don't lose the combo if only a modifier is pressed
                last = self.last_pressed
                if key.key != ExtKey.No:
                    repress = self._current_map is not None
                    current.deactivate()
                    self.initialize()
                    if repress:
                        # if key combo is initialized half-way, repress the last key
                        return self._press_one_key(key)

                if last is not None and last.key == ExtKey.No and key.key != ExtKey.No:
                    return self._press_one_key(key)

                return False
            _is_parametric = True

        if isinstance(callback, QtKeyMap):
            callback._obj = self._obj
            try:
                self.activate(key)
            except TriggerParent:
                return False
        else:
            current.deactivate()
            self.initialize()
            if self._obj is not None:
                callback = callback.__get__(self._obj)
            try:
                if _is_parametric:
                    callback(key.key_string())
                else:
                    callback()
            except TriggerParent:
                return False

        last = self.last_pressed
        if last is not None and last.key == ExtKey.No and last.modifier == key.modifier:
            self._current_combo.pop()
        self._current_combo.append(key)
        return True

    def activate(self, key: KeyType) -> None:
        """Activate key combo trigger."""
        key = QtKeys(key)

        old = self.current_map
        current = self.current_map[key]
        if not isinstance(current, QtKeyMap):
            self.initialize()
            raise KeyError(f"Key {key} is not a child")
        self._current_map = current
        if current._obj is None:
            current._activated_callback()
        else:
            current._activated_callback(current._obj)
        old.deactivate()
        return None

    def deactivate(self) -> None:
        """Deactivate key combo trigger."""
        if self._obj is None:
            self._deactivated_callback()
        else:
            self._deactivated_callback(self._obj)
        return None

    def initialize(self):
        self._current_map = None
        self._current_combo.clear()
        return None

    def to_widget(self):
        from ._keymapview import QtKeyMapView

        return QtKeyMapView.from_keymap(self)

    def __setitem__(self, key, value):
        if callable(value):
            value = BoundCallback(value)
        elif not isinstance(value, QtKeyMap):
            raise TypeError("Values in QtKeyMap must be QtKeyMap or callable.")
        self._keymap[QtKeys(key)] = value

    def __delitem__(self, key):
        del self._keymap[QtKeys(key)]

    def __iter__(self):
        return iter(self._keymap)

    def __len__(self):
        return len(self._keymap)


def _DUMMY_CALLBACK(*args):
    """Dummy callback function."""


_DUMMY_CALLBACK = BoundCallback(_DUMMY_CALLBACK)



from typing import TYPE_CHECKING
from scipy import optimize as sp_opt
import numpy as np
import pandas as pd
from magicgui.widgets import Container, ComboBox, PushButton, SpinBox

from tabulous._magicgui import find_current_table, SelectionWidget

if TYPE_CHECKING:
    from tabulous._qt._table._base import QMutableSimpleTable


class OptimizerWidget(Container):
    """A widget for Excel's solver-like optimization."""

    _current_widget = None

    def __init__(self):
        self._cost_selector = SelectionWidget(
            label="Cost", tooltip="Select the cost cell"
        )
        self._param_selector = SelectionWidget(
            label="Parameters", tooltip="Select the parameter cells"
        )
        self._minimize_cbox = ComboBox(
            label="Optimize by", choices=["minimize", "maximize"]
        )
        self._maxiter = SpinBox(value=50, label="Max Iterations")

        self._call_button = PushButton(text="Run", tooltip="Run optimization")
        self._call_button.changed.connect(self._on_called)
        super().__init__(
            widgets=[
                self._cost_selector,
                self._param_selector,
                self._minimize_cbox,
                self._maxiter,
                self._call_button,
            ]
        )

    def _on_called(self):
        table = find_current_table(self)
        df = table.data_shown
        dst = self._cost_selector.value.as_iat(df)
        params = self._param_selector.value.as_iloc_slices(df)
        qtable = table._qwidget
        if dst not in qtable._qtable_view._table_map:
            raise ValueError(f"{dst} has no reference.")

        if self._minimize_cbox.value == "minimize":
            cost_function = _get_minimize_target(qtable, dst, params)
        else:
            cost_function = _get_maximize_target(qtable, dst, params)

        data = qtable.dataShown().iloc[params]
        param0 = data.to_numpy(dtype=np.float64, copy=False)

        def callback(x):
            return qtable.refreshTable(True)

        with qtable._mgr.merging(lambda x: "optimize"):
            sp_opt.minimize(
                cost_function,
                param0,
                callback=callback,
                options=dict(maxiter=self._maxiter.value),
            )
        return None

    @classmethod
    def new(cls) -> "OptimizerWidget":
        if cls._current_widget is None:
            cls._current_widget = cls()
        return cls._current_widget


def _get_minimize_target(
    table: "QMutableSimpleTable", dst: tuple[int, int], params: tuple[slice, slice]
):
    def cost_func(p):
        table.setDataFrameValue(*params, pd.DataFrame(p))
        val = table.dataShown().iat[dst]
        return float(val)

    return cost_func


def _get_maximize_target(
    table: "QMutableSimpleTable", dst: tuple[int, int], params: tuple[slice, slice]
):
    def cost_func(p):
        table.setDataFrameValue(*params, pd.DataFrame(p))
        val = table.dataShown().iat[dst]
        return -float(val)

    return cost_func



from __future__ import annotations
from typing import TYPE_CHECKING, Iterator, cast
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt, Signal

from tabulous._qt._action_registry import QActionRegistry
from tabulous._qt._proxy_button import HeaderAnchorMixin

if TYPE_CHECKING:
    from ._enhanced_table import _QTableViewEnhanced


class QDataFrameHeaderView(QtW.QHeaderView, QActionRegistry[int]):
    """The header view for the tabulous tables."""

    _Orientation: Qt.Orientation
    selectionChangedSignal = Signal(int, int)

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        QtW.QHeaderView.__init__(self, self._Orientation, parent)
        QActionRegistry.__init__(self)
        self.setSelectionMode(QtW.QHeaderView.SelectionMode.SingleSelection)
        self.setSectionsClickable(True)
        self.sectionPressed.connect(self._on_section_pressed)  # pressed
        self.sectionClicked.connect(self._on_section_clicked)  # released
        self.sectionEntered.connect(self._on_section_entered)  # dragged
        self.sectionResized.connect(self._on_section_resized)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)

        self._header_widgets: dict[int, QtW.QWidget] = {}

    # fmt: off
    if TYPE_CHECKING:
        def parentWidget(self) -> _QTableViewEnhanced: ...
    # fmt: on

    @property
    def selection_model(self):
        return self.parentWidget()._selection_model

    def _on_section_pressed(self, logicalIndex: int) -> None:
        self.selection_model.jump_to(*self._index_for_selection_model(logicalIndex))
        self.selection_model.set_shift(True)
        return None

    def _on_section_entered(self, logicalIndex: int) -> None:
        self.selection_model.move_to(*self._index_for_selection_model(logicalIndex))
        return None

    def _on_section_clicked(self, logicalIndex) -> None:
        self.selection_model.set_shift(False)

    def _on_section_resized(
        self, logicalIndex: int, oldSize: int, newSize: int
    ) -> None:
        for idx, widget in self._header_widgets.items():
            if idx < logicalIndex:
                continue
            widget.move(widget.x() + newSize - oldSize, widget.y())

    def _show_context_menu(self, pos: QtCore.QPoint) -> None:
        index = self.logicalIndexAt(pos)
        # press header if it is not selected.
        for sel in self._iter_selections():
            if sel.start <= index < sel.stop:
                break
        else:
            self._on_section_pressed(index)
        return self.execContextMenu(self.viewport().mapToGlobal(pos), index)

    def _iter_selections(self) -> Iterator[slice]:
        """Iterate selections"""
        raise NotImplementedError()

    def _index_for_selection_model(self, logicalIndex: int) -> tuple[int, int]:
        raise NotImplementedError()

    def visualRectAtIndex(self, index: int) -> QtCore.QRect:
        """Return the visual rect of the given index."""
        raise NotImplementedError()

    @staticmethod
    def drawBorder(painter: QtGui.QPainter, rect: QtCore.QRect):
        """Draw the opened border of a section."""
        raise NotImplementedError()

    def drawCurrent(self, painter: QtGui.QPainter, rect: QtCore.QRect):
        """Draw the current index if exists."""
        raise NotImplementedError()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        super().paintEvent(event)
        painter = QtGui.QPainter(self.viewport())
        color = self.parentWidget()._get_selection_color()
        pen = QtGui.QPen(color, 3)
        painter.setPen(pen)

        # paint selections
        for _slice in self._iter_selections():
            rect_start = self.visualRectAtIndex(_slice.start)
            rect_stop = self.visualRectAtIndex(_slice.stop - 1)
            rect = rect_start | rect_stop
            self.drawBorder(painter, rect)

        # paint current
        self.drawCurrent(painter)
        return None

    def mouseReleaseEvent(self, e: QtGui.QMouseEvent) -> None:
        if editor := self.parentWidget()._focused_widget:
            editor.setFocus()
        return super().mouseReleaseEvent(e)

    def sectionWidget(self, idx: int) -> QtW.QWidget | None:
        """The widget anchored at the given section."""
        return self._header_widgets.get(idx, None)

    def setSectionWidget(self, idx: int, widget: QtW.QWidget) -> None:
        """Set the widget anchored at the given section."""
        if not isinstance(widget, QtW.QWidget):
            raise TypeError(f"Expected a QWidget, got {type(widget)}")
        if idx in self._header_widgets:
            self.removeSectionWidget(idx)
        self._header_widgets[idx] = widget
        widget.setParent(self.viewport())
        w, h = widget.width(), widget.height()
        rect = self.visualRectAtIndex(idx)
        rect.adjust(2, 2, -2, -2)
        rect.setLeft(rect.left() + rect.width() - w)
        rect.setTop(rect.top() + rect.height() - h)
        widget.setGeometry(rect)
        if isinstance(widget, HeaderAnchorMixin):
            _widget = cast(HeaderAnchorMixin, widget)
            _widget.on_installed(self.parentWidget().parentTable(), idx)
        widget.show()
        return None

    def removeSectionWidget(self, idx: int | None = None):
        table = self.parentWidget().parentTable()
        if idx is None:
            for idx in self._header_widgets:
                widget = self._header_widgets[idx]
                widget.hide()
                if isinstance(widget, HeaderAnchorMixin):
                    _widget = cast(HeaderAnchorMixin, widget)
                    _widget.on_uninstalled(table, idx)
                widget.setParent(None)
            self._header_widgets.clear()
        else:
            widget = self._header_widgets.pop(idx)
            widget.hide()
            if isinstance(widget, HeaderAnchorMixin):
                _widget = cast(HeaderAnchorMixin, widget)
                _widget.on_uninstalled(table, idx)
            widget.setParent(None)
        return None


class QHorizontalHeaderView(QDataFrameHeaderView):
    _Orientation = Qt.Orientation.Horizontal

    def visualRectAtIndex(self, index: int) -> QtCore.QRect:
        x = self.sectionViewportPosition(index)
        y = self.rect().top()
        height = self.height()
        width = self.sectionSize(index)
        return QtCore.QRect(x, y, width, height)

    @staticmethod
    def drawBorder(painter: QtGui.QPainter, rect: QtCore.QRect):
        return painter.drawPolyline(
            rect.bottomLeft(),
            rect.topLeft(),
            rect.topRight(),
            rect.bottomRight(),
        )

    def drawCurrent(self, painter: QtGui.QPainter):
        row, col = self.selection_model.current_index
        if row < 0 and col >= 0:
            rect_current = self.visualRectAtIndex(col)
            rect_current.adjust(1, 1, -1, -1)
            color = self.parentWidget()._get_current_index_color()
            pen = QtGui.QPen(color, 3)
            painter.setPen(pen)
            painter.drawRect(rect_current)
        return None

    def _iter_selections(self):
        yield from self.selection_model.iter_col_selections()

    def _index_for_selection_model(self, logicalIndex):
        return -1, logicalIndex


class QVerticalHeaderView(QDataFrameHeaderView):
    _Orientation = Qt.Orientation.Vertical

    def visualRectAtIndex(self, index: int) -> QtCore.QRect:
        x = self.rect().left()
        y = self.sectionViewportPosition(index)
        height = self.sectionSize(index)
        width = self.width()
        return QtCore.QRect(x, y, width, height)

    @staticmethod
    def drawBorder(painter: QtGui.QPainter, rect: QtCore.QRect):
        return painter.drawPolyline(
            rect.topRight(),
            rect.topLeft(),
            rect.bottomLeft(),
            rect.bottomRight(),
        )

    def drawCurrent(self, painter: QtGui.QPainter):
        row, col = self.selection_model.current_index
        if col < 0 and row >= 0:
            rect_current = self.visualRectAtIndex(row)
            rect_current.adjust(1, 1, -1, -1)
            color = self.parentWidget()._get_current_index_color()
            pen = QtGui.QPen(color, 4)
            painter.setPen(pen)
            painter.drawRect(rect_current)
        return None

    def _iter_selections(self):
        yield from self.selection_model.iter_row_selections()

    def _index_for_selection_model(self, logicalIndex):
        return logicalIndex, -1



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path
import weakref
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, QEvent, QTimer

from ._base import _QtMainWidgetBase
from tabulous._keymap import QtKeyMap
from tabulous.types import TabPosition
from tabulous._utils import get_config

if TYPE_CHECKING:
    from tabulous.widgets import TableViewer
    from tabulous._qt._table_stack import QTabbedTableStack
    from tabulous._qt._dockwidget import QtDockWidget

ICON_DIR = Path(__file__).parent.parent / "_icons"


class QMainWidget(QtW.QSplitter, _QtMainWidgetBase):
    _keymap = QtKeyMap()

    def __init__(self, tab_position: TabPosition | str = TabPosition.top):
        QtW.QSplitter.__init__(self)
        _QtMainWidgetBase.__init__(self, tab_position)
        self.setOrientation(Qt.Orientation.Vertical)
        self._toolbar = None
        self._statusbar = QRichStatusBar(self)
        self._statusbar.setFixedHeight(25)
        self.addWidget(self._statusbar)

        cfg = get_config()
        if cfg.window.nonmain_style:
            # set style
            self.applyTheme(cfg.window.theme)
            self.updateWidgetStyle()

    def setCentralWidget(self, wdt: QTabbedTableStack):
        """Mimicking QMainWindow's method by adding a widget to the layout."""
        self.addWidget(wdt)
        wdt.setMinimumSize(200, 200)
        self.setMinimumSize(200, 200)
        return None

    def toolBarVisible(self) -> bool:
        """Visibility of toolbar"""
        if self._toolbar is None:
            return False
        else:
            return self._toolbar.isVisible()

    def setToolBarVisible(self, visible: bool):
        """Set visibility of toolbar"""
        if visible and self._toolbar is None:
            from tabulous._qt._toolbar import QTableStackToolBar

            self._toolbar = QTableStackToolBar(self)
            self.insertWidget(0, self._toolbar)
            self.updateWidgetStyle()
            self._toolbar.sliceChanged.connect(self.setTableSelection)

        return self._toolbar.setVisible(visible)

    def consoleVisible(self) -> bool:
        """True if embeded console is visible."""
        if self._console_widget is None:
            return False
        else:
            return self._console_widget.isVisible()

    def setConsoleVisible(self, visible: bool) -> None:
        """Set visibility of embeded console widget."""
        if visible and self._console_widget is None:
            from tabulous._qt._console import QtConsole

            qtconsole = QtConsole()
            qtconsole.connect_parent(self._table_viewer)
            self.addWidget(qtconsole)
            self._console_widget = qtconsole

            if qtconsole.shell is not None:
                qtconsole.update_console(self._queued_ns)
                self._queued_ns.clear()

        self._console_widget.setVisible(visible)

        if visible:
            self._console_widget.setFocus()
        else:
            self.setCellFocus()

    def statusBar(self) -> QRichStatusBar:
        return self._statusbar


_REORDER_INSTANCES = frozenset({QEvent.Type.WindowActivate, QEvent.Type.ZOrderChange})

_HIDE_TOOLTIPS = frozenset(
    {
        QEvent.Type.MouseButtonPress,
        QEvent.Type.MouseButtonDblClick,
        QEvent.Type.KeyPress,
        QEvent.Type.Move,
        QEvent.Type.Resize,
        QEvent.Type.Show,
        QEvent.Type.Hide,
        QEvent.Type.Close,
        QEvent.Type.WindowStateChange,
        QEvent.Type.WindowDeactivate,
        QEvent.Type.FocusOut,
        QEvent.Type.ZOrderChange,
    }
)

# Flag to assert that the new version of tabulous will be notified only once
_NOTIFIED = False


class QMainWindow(QtW.QMainWindow, _QtMainWidgetBase):
    _instances: list[QMainWindow] = []
    _keymap = QtKeyMap()
    _dock_widgets: weakref.WeakValueDictionary[str, QtDockWidget]
    _is_mainwindow = True

    def __init__(
        self,
        tab_position: TabPosition | str = TabPosition.top,
    ):
        from tabulous._qt._toolbar import QTableStackToolBar

        cfg = get_config()

        super().__init__()
        _QtMainWidgetBase.__init__(self, tab_position=tab_position)
        self.setWindowTitle("tabulous")

        self.setWindowIcon(QtGui.QIcon(str(ICON_DIR / "window_icon.png")))

        self._console_dock_widget = None
        self._dock_widgets = weakref.WeakValueDictionary()

        # ask if it is OK to close
        self._ask_on_close = cfg.window.ask_on_close

        self._toolbar = QTableStackToolBar(self)
        self.addToolBar(self._toolbar)
        self._toolbar.setMovable(False)  # nested toolbar causes layout problems
        self._toolbar.sliceChanged.connect(self.setTableSelection)

        self._tablestack.setMinimumSize(400, 250)
        self.resize(800, 600)
        self.setStatusBar(QRichStatusBar(self))
        QMainWindow._instances.append(self)

        if cfg.window.show_console:
            self.setConsoleVisible(True)

        # set style
        self.applyTheme(cfg.window.theme)
        self.updateWidgetStyle()

    def consoleVisible(self) -> bool:
        """True if embeded console is visible."""
        if self._console_widget is None:
            return False
        else:
            return self._console_widget.isVisible()

    def setConsoleVisible(self, visible: bool) -> None:
        """Set visibility of embeded console widget."""
        if visible and self._console_widget is None:
            from tabulous._qt._console import QtConsole

            qtconsole = QtConsole()
            qtconsole.connect_parent(self._table_viewer)
            dock = self.addDockWidget(qtconsole, name="Console", area="bottom")
            qtconsole.setDockParent(dock)
            dock.setSourceObject(qtconsole)
            self._console_widget = qtconsole

            if qtconsole.shell is not None:
                qtconsole.update_console(self._queued_ns)
                self._queued_ns.clear()

            qtconsole.update_theme(self._style_theme)
        else:
            dock = self._console_widget.dockParent()

        dock.setVisible(visible)

        if visible:
            if dock.isFloating():
                QTimer.singleShot(0, dock.activateWindow)
            self._console_widget.setFocus()
        else:
            self.setCellFocus()

    def addDockWidget(
        self,
        qwidget: QtW.QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] = None,
    ):
        from .._dockwidget import QtDockWidget

        name = name or qwidget.objectName()
        dock = QtDockWidget(
            self,
            qwidget,
            name=name.replace("_", " "),
            area=area,
            allowed_areas=allowed_areas,
        )

        super().addDockWidget(QtDockWidget.areas[area], dock)
        return dock

    @classmethod
    def currentViewer(cls) -> TableViewer:
        """Return the current TableViewer widget."""
        window = cls._instances[-1] if cls._instances else None
        return window._table_viewer if window else None

    def show(self) -> None:
        global _NOTIFIED

        super().show()

        if _NOTIFIED:
            return
        _NOTIFIED = True
        from tabulous._fetch_and_install import get_worker

        worker = get_worker()

        @worker.returned.connect
        def _(v: str):
            if v is not None and get_config().window.notify_latest:
                self._tablestack.notifyLatestVersion(v)
            # NOTE: uncomment the following line to test the notification
            # self._tablestack.notifyLatestVersion("0.X.X")

        worker.start()

    def close(self, ask: bool | None = False) -> bool:
        if ask is not None:
            self._ask_on_close = ask
        return super().close()

    def event(self, e: QEvent):
        type = e.type()
        if type == QEvent.Type.Close:
            if self._ask_on_close and not self._tablestack.isEmpty():
                msgbox = QtW.QMessageBox(self)
                msgbox.setWindowTitle("tabulous")
                msgbox.setIcon(QtW.QMessageBox.Icon.Question)
                msgbox.setText("Are you sure to close this window?")
                btn = msgbox.addButton(QtW.QMessageBox.StandardButton.Yes)
                msgbox.addButton(QtW.QMessageBox.StandardButton.No)
                btn.setShortcut(QtGui.QKeySequence("Ctrl+W"))

                cbox = QtW.QCheckBox("Don't ask again")
                msgbox.setCheckBox(cbox)
                btn = msgbox.exec()
                if cbox.isChecked():
                    get_config().window.ask_on_close = False
                if btn == QtW.QMessageBox.StandardButton.No:
                    e.ignore()
                    return True
            # when we close the MainWindow, remove it from the instances list
            try:
                QMainWindow._instances.remove(self)
            except ValueError:
                pass
            get_config().as_toml()  # save config

        elif type in _REORDER_INSTANCES:
            # upon activation or raise_, put window at the end of _instances
            try:
                inst = QMainWindow._instances
                inst.append(inst.pop(inst.index(self)))
            except ValueError:
                pass

        elif type in _HIDE_TOOLTIPS:
            self._toolbar.hideTabTooltips()
            self._toolbar.currentToolBar().hideTabTooltips()

        return super().event(e)

    def toolBarVisible(self) -> bool:
        """Visibility of toolbar"""
        return self._toolbar.isVisible()

    def setToolBarVisible(self, visible: bool):
        """Set visibility of toolbar"""
        return self._toolbar.setVisible(visible)

    @classmethod
    def reload_config(cls):
        """Reload tabulous config and apply to all TableViewer instances"""
        cfg = get_config()
        for self in cls._instances:
            self.applyTheme(cfg.window.theme)
            if toolbar := self._toolbar:
                if not self._tablestack.isEmpty():
                    toolbar._corner_widget.setVisible(
                        get_config().window.selection_editor
                    )
            for table in self._table_viewer.tables:
                table._qwidget._qtable_view.load_config(cfg)
        return None


class QRichStatusBar(QtW.QStatusBar):
    """Custom status bar with rich text support."""

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self._label = QtW.QLabel("")
        self._label.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        self._label.setFrameStyle(QtW.QFrame.Shape.NoFrame)
        self.addWidget(self._label)
        self.setContentsMargins(0, 0, 0, 0)

    def showMessage(self, msg: str, timeout: int = 0) -> None:
        self._label.resize(self.width() - 32, self._label.height())
        return self._label.setText(msg)



from __future__ import annotations
from typing import (
    Any,
    Callable,
    Iterable,
    NewType,
    Tuple,
    List,
    Union,
    TYPE_CHECKING,
    NamedTuple,
    SupportsIndex,
    Mapping,
    MutableSequence,
)
from enum import Enum

import numpy as np
from numpy.typing import ArrayLike, NDArray

if TYPE_CHECKING:
    import pandas as pd

    _TableLike = Union[pd.DataFrame, dict, Iterable, ArrayLike]
    from .widgets._component import SelectionRanges
else:
    _TableLike = Any
    SelectionRanges = MutableSequence[Tuple[slice, slice]]


__all__ = [
    "TableData",
    "TableColumn",
    "TableDataTuple",
    "TableInfo",
    "ItemInfo",
    "HeaderInfo",
    "SelectionRanges",
    "SelectionType",
]

if TYPE_CHECKING:
    TableData = NewType("TableData", pd.DataFrame)
    TableColumn = NewType("TableColumn", pd.Series)
else:
    TableData = NewType("TableData", Any)
    TableColumn = NewType("TableColumn", Any)

TableDataTuple = NewType("TableDataTuple", tuple)


class TableInfoAlias(type):
    def __getitem__(cls, names: str | tuple[str, ...]):
        from typing_extensions import Annotated

        if isinstance(names, str):
            names = (names,)
        else:
            for name in names:
                if not isinstance(name, str):
                    raise ValueError(
                        f"Cannot subscribe type {type(name)} to TableInfo."
                    )

        return Annotated[TableInfoInstance, {"column_choice_names": names}]


class TableInfo(metaclass=TableInfoAlias):
    """
    A generic type to describe a DataFrame and its column names.

    ``TableInfo["x", "y"]`` is equivalent to ``tuple[pd.DataFrame, str, str]``
    with additional information for magicgui construction.
    """

    def __new__(cls, *args, **kwargs):
        raise TypeError(f"Type {cls.__name__} cannot be instantiated.")


class TableInfoInstance(Tuple["pd.DataFrame", List[str]]):
    def __new__(cls, *args, **kwargs):
        raise TypeError(f"Type {cls.__name__} cannot be instantiated.")


_IntArray = NDArray[np.integer]
_BoolArray = NDArray[np.bool_]
_IntOrBoolArray = Union[_IntArray, _BoolArray]
ProxyType = Union[Callable[["pd.DataFrame"], _IntOrBoolArray], _IntOrBoolArray]


class TabPosition(Enum):
    """Enum for tab position."""

    top = "top"
    left = "left"
    bottom = "bottom"
    right = "right"


class _InfoVar:
    def __init__(self, name: str) -> None:
        self._name = name

    def __repr__(self) -> str:
        return self._name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, _InfoVar):
            return self._name == other._name
        return False


class ItemInfo(NamedTuple):
    """
    A named tuple for item update.

    Value takes (row, column, value, old_value) where

    row : int or slice
        Row index or slice where item edited.
    column : int or slice
        Column index or slice where item edited.
    value : Any
        New value of the item.
    old_value : Any
        Old value of the item.

    If a row or column is deleted, the value is set to ``DELETED``.
    If a row or column is inserted, the old_value is set to ``INSERTED``.
    """

    # class variables
    DELETED = _InfoVar("DELETED")
    INSERTED = _InfoVar("INSERTED")

    row: int | slice
    column: int | slice
    value: Any
    old_value: Any

    @property
    def col(self) -> int | slice:
        """Alias of `column`."""
        return self.column


class HeaderInfo(NamedTuple):
    """
    A named tuple for header update.

    Value takes (index, value, old_value) where

    index : int
        Index where item edited.
    value : Any
        New value of the item.
    old_value : Any
        Old value of the item.
    """

    index: int
    value: Any
    old_value: Any


class EvalInfo(NamedTuple):
    """
    A named tuple for evaluation.

    Value takes (row, column, expr, old_value, is_ref) where

    pos : tuple[int, int]
        The visual position of the cell where expression is evaluated.
    source_pos : tuple[int, int]
        The source position of the cell where expression is evaluated.
    expr : str
        Expression to be evaluated.
    is_ref: bool
        Whether the expression take references in the table.
    """

    pos: tuple[int, int]
    source_pos: tuple[int, int]
    expr: str
    is_ref: bool

    @property
    def col(self) -> int:
        """Alias of `column`."""
        return self.column


_Sliceable = Union[SupportsIndex, slice]
_SingleSelection = Tuple[_Sliceable, _Sliceable]
SelectionType = List[_SingleSelection]


def __getattr__(name: str) -> Any:
    if name == "FilterType":
        import warnings

        warnings.warn(
            "'FilterType 'is deprecated. Please use 'ProxyType' instead.",
            DeprecationWarning,
        )
        return ProxyType
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


ColorType = Union[str, Iterable[int]]
ColorMapping = Union[Callable[[Any], ColorType], Mapping[str, ColorType]]



from __future__ import annotations
from typing import Callable
import numpy as np
import pandas as pd

ORD_A = ord("A")
CHARS = list("ABCDEFGHIJKLMNOPQRSTUVWXYZ") + [""]
CHARS_SET = set(CHARS)
LONGEST = np.array(CHARS[:-1], dtype=object)


def str_to_num(s: str):
    out = 0
    for i, c in enumerate(reversed(s)):
        if c not in CHARS_SET:
            raise ValueError(f"Character {c} is not allowed.")
        out += (ord(c) - ORD_A) ** (i + 1)
    return out


def _iter_char(start: int, stop: int):
    if stop >= 26**4:
        raise ValueError("Stop must be less than 26**4 - 1")
    base_repr = np.base_repr(start, 26)
    current = np.zeros(4, dtype=np.int8)
    offset = 4 - len(base_repr)
    for i, d in enumerate(base_repr):
        current[i + offset] = int(d, 26)

    current[:3] -= 1
    for _ in range(start, stop):
        yield "".join(CHARS[s] for s in current)
        current[3] += 1
        for i in [3, 2, 1]:
            if current[i] >= 26:
                over = current[i] - 25
                current[i] = 0
                current[i - 1] += over


def char_arange(start: int, stop: int = None) -> np.ndarray:
    """
    A char version of np.arange.

    Examples
    --------
    >>> char_arange(3)  # array(["A", "B", "C"])
    >>> char_arange(25, 28)  # array(["Z", "AA", "AB"])
    """
    global LONGEST
    if stop is None:
        start, stop = 0, start
    nmax = len(LONGEST)
    if stop <= nmax:
        return np.array(LONGEST[start:stop], dtype=object)
    LONGEST = np.append(LONGEST, np.fromiter(_iter_char(nmax, stop), dtype=object))
    return LONGEST[start:].copy()


class UniqueName:
    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"UniqueName({self.name!r})"

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other


HEADER = UniqueName("header")


def char_range_index(stop: int) -> pd.Index:
    return pd.Index(char_arange(stop), name=HEADER)


def is_ranged(index: pd.Index) -> bool:
    """Check if given index is range-like."""
    return isinstance(index, pd.RangeIndex) or index.name is HEADER


def as_not_ranged(index: pd.Index) -> None:
    if index.name is HEADER:
        index.name = None


def as_constructor(index: pd.Index) -> Callable[[int], pd.Index] | None:
    if isinstance(index, pd.RangeIndex):
        return pd.RangeIndex
    elif index.name is HEADER:
        return char_range_index
    else:
        return None



from __future__ import annotations
from pathlib import Path
from enum import Enum
from typing import Callable, Literal, overload
from qtpy import QtWidgets as QtW, QtCore
from tabulous._utils import load_file_open_path

# Modified from magicgui
# See https://github.com/pyapp-kit/magicgui/blob/main/magicgui/backends/_qtpy/widgets.py
class FileDialogMode(Enum):
    """FileDialog mode options."""

    EXISTING_FILE = "r"
    EXISTING_FILES = "rm"
    OPTIONAL_FILE = "w"
    EXISTING_DIRECTORY = "d"


QFILE_DIALOG_MODES: dict[FileDialogMode, Callable] = {
    FileDialogMode.EXISTING_FILE: QtW.QFileDialog.getOpenFileName,
    FileDialogMode.EXISTING_FILES: QtW.QFileDialog.getOpenFileNames,
    FileDialogMode.OPTIONAL_FILE: QtW.QFileDialog.getSaveFileName,
    FileDialogMode.EXISTING_DIRECTORY: QtW.QFileDialog.getExistingDirectory,
}


class QtFileHistoryManager(QtCore.QObject):
    def __init__(
        self,
        default_path: Path | str | bytes | None = None,
    ):
        super().__init__()
        self._hist: list[Path] = [Path(path.strip()) for path in load_file_open_path()]
        self.default_path = default_path or Path.cwd()
        self._instances: dict[int, QtFileHistoryManager] = {}
        self._parent_class = None

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        _id = id(obj)
        if (out := self._instances.get(_id, None)) is None:
            out = self.__class__(default_path=self.default_path)
            if self._parent_class is not None and issubclass(
                self._parent_class, QtCore.QObject
            ):
                out.setParent(obj)
            self._instances[_id] = out
        return out

    def __set_name__(self, owner: type | None, name: str):
        self._parent_class = owner

    def recentlyVisitedDirectory(self) -> Path:
        """Return the mose recently visited, existing directory."""
        for path in reversed(self._hist):
            if path.exists():
                return path

        return self.default_path

    @overload
    def openFileDialog(
        self,
        mode: Literal[
            "r",
            "w",
            "d",
            FileDialogMode.EXISTING_FILE,
            FileDialogMode.OPTIONAL_FILE,
            FileDialogMode.EXISTING_DIRECTORY,
        ],
        caption: str | None = None,
        filter: str | None = None,
    ) -> Path | None:
        ...

    @overload
    def openFileDialog(
        self,
        mode: Literal["rm"] | Literal[FileDialogMode.EXISTING_FILES],
        caption: str | None = None,
        filter: str | None = None,
    ) -> list[Path] | None:
        ...

    def openFileDialog(
        self,
        mode=FileDialogMode.EXISTING_FILE,
        caption=None,
        filter=None,
    ):
        if not isinstance(parent := self.parent(), QtW.QWidget):
            parent = None

        directory = str(self.recentlyVisitedDirectory().parent)

        out = QFILE_DIALOG_MODES[FileDialogMode(mode)](
            parent=parent,
            caption=caption,
            directory=directory,
            filter=filter,
        )
        if out is None:
            return None

        if mode is FileDialogMode.EXISTING_DIRECTORY:
            path = out
        else:
            path, _ = out

        if isinstance(path, str):
            if path != "":
                out = Path(path)
                if out.is_file():
                    self._hist.append(out.parent)
                else:
                    self._hist.append(out)
            else:
                out = None
        elif isinstance(path, list):
            if len(path) > 0:
                out = [Path(p) for p in path]
                self._hist.append(out[-1])
            else:
                out = []
        else:
            out = None
        return out

    @classmethod
    def requestPath(
        cls,
        mode=FileDialogMode.EXISTING_FILE,
        caption=None,
        filter=None,
    ):
        self = cls()
        return self.openFileDialog(mode, caption, filter)

    def clearHistory(self):
        """Clear the history of visited directories."""
        return self._hist.clear()

    def defaultPath(self) -> Path:
        """Get the default path."""
        return self._default_path

    def setDefaultPath(self, path: Path | str | bytes):
        """Set the default path."""
        path = Path(path)
        if not path.exists():
            raise ValueError(f"Path {path!r} does not exist.")
        self._default_path = path



from typing import Callable
from ._table_base import QBaseTable, QMutableTable
from tabulous.exceptions import TriggerParent


@QBaseTable._keymap.bind("Up", dr=-1, dc=0)
@QBaseTable._keymap.bind("Down", dr=1, dc=0)
@QBaseTable._keymap.bind("Left", dr=0, dc=-1)
@QBaseTable._keymap.bind("Right", dr=0, dc=1)
@QBaseTable._keymap.bind("PageUp", dr=-10, dc=0)
@QBaseTable._keymap.bind("PageDown", dr=10, dc=0)
@QBaseTable._keymap.bind("Home", dr=0, dc=-10)
@QBaseTable._keymap.bind("End", dr=0, dc=10)
def _(self: QBaseTable, dr, dc):
    return self._qtable_view._selection_model.move(dr, dc, allow_header=True)


@QBaseTable._keymap.bind("Shift+Up", dr=-1, dc=0)
@QBaseTable._keymap.bind("Shift+Down", dr=1, dc=0)
@QBaseTable._keymap.bind("Shift+Left", dr=0, dc=-1)
@QBaseTable._keymap.bind("Shift+Right", dr=0, dc=1)
@QBaseTable._keymap.bind("Shift+PageUp", dr=-10, dc=0)
@QBaseTable._keymap.bind("Shift+PageDown", dr=10, dc=0)
@QBaseTable._keymap.bind("Shift+Home", dr=0, dc=-10)
@QBaseTable._keymap.bind("Shift+End", dr=0, dc=10)
def _(self: QBaseTable, dr, dc):
    self._qtable_view._selection_model.set_shift(True)
    return self._qtable_view._selection_model.move(dr, dc)


@QBaseTable._keymap.bind("Ctrl+Up", row=0, column=None)
@QBaseTable._keymap.bind("Ctrl+Down", row=-1, column=None)
@QBaseTable._keymap.bind("Ctrl+Left", row=None, column=0)
@QBaseTable._keymap.bind("Ctrl+Right", row=None, column=-1)
@QBaseTable._keymap.bind("Ctrl+Shift+Up", row=0, column=None)
@QBaseTable._keymap.bind("Ctrl+Shift+Down", row=-1, column=None)
@QBaseTable._keymap.bind("Ctrl+Shift+Left", row=None, column=0)
@QBaseTable._keymap.bind("Ctrl+Shift+Right", row=None, column=-1)
def _(self: QBaseTable, row, column):
    return self.moveToItem(row, column, clear_selection=False)


@QBaseTable._keymap.bind("Ctrl+Alt+Up", dr=-1, dc=0)
@QBaseTable._keymap.bind("Ctrl+Alt+Down", dr=1, dc=0)
@QBaseTable._keymap.bind("Ctrl+Alt+Left", dr=0, dc=-1)
@QBaseTable._keymap.bind("Ctrl+Alt+Right", dr=0, dc=1)
@QBaseTable._keymap.bind("Ctrl+Alt+PageUp", dr=-10, dc=0)
@QBaseTable._keymap.bind("Ctrl+Alt+PageDown", dr=10, dc=0)
@QBaseTable._keymap.bind("Ctrl+Alt+Home", dr=0, dc=-10)
@QBaseTable._keymap.bind("Ctrl+Alt+End", dr=0, dc=10)
def _(self: QBaseTable, dr, dc):
    """Scroll without moving the selection."""
    vbar = self._qtable_view.verticalScrollBar()
    hbar = self._qtable_view.horizontalScrollBar()
    dv = dr * 75 * self._qtable_view._zoom
    dh = dc * 75 * self._qtable_view._zoom
    vbar.setValue(max(vbar.minimum(), min(vbar.maximum(), vbar.value() + dv)))
    hbar.setValue(max(hbar.minimum(), min(hbar.maximum(), hbar.value() + dh)))


# Following key combos are only available when the table is not in a viewer
def _check_no_viewer(func: Callable[[QBaseTable], None]):
    def _wrapped(self: QBaseTable, **kwargs):
        if self.parentViewer() is None:
            func(self, **kwargs)
        else:
            raise TriggerParent

    return _wrapped


QBaseTable._keymap.bind("Ctrl+C", headers=False)(
    _check_no_viewer(QBaseTable.copyToClipboard)
)
QBaseTable._keymap.bind("Ctrl+C, Ctrl+H", headers=True)(
    _check_no_viewer(QBaseTable.copyToClipboard)
)
QMutableTable._keymap.bind("Ctrl+V")(_check_no_viewer(QMutableTable.pasteFromClipBoard))
QBaseTable._keymap.bind("Ctrl+A")(
    _check_no_viewer(lambda self: self._qtable_view.selectAll)
)
QMutableTable._keymap.bind("Delete")(_check_no_viewer(QMutableTable.deleteValues))
QMutableTable._keymap.bind("Backspace")(_check_no_viewer(QMutableTable.deleteValues))

QBaseTable._keymap.bind("Ctrl+K")(_check_no_viewer(QBaseTable.setFocus))
QBaseTable._keymap.bind("Ctrl+H")(_check_no_viewer(QBaseTable.undoStackView))
QBaseTable._keymap.bind("Menu")(_check_no_viewer(QBaseTable.showContextMenuAtIndex))
QBaseTable._keymap.bind("F6")(_check_no_viewer(QBaseTable.raiseSlotError))
QMutableTable._keymap.bind("Ctrl+Z")(_check_no_viewer(QMutableTable.undo))
QMutableTable._keymap.bind("Ctrl+Y")(_check_no_viewer(QMutableTable.redo))


@QBaseTable._keymap.bind("Ctrl+K, E")
@_check_no_viewer
def _(self: QBaseTable):
    try:
        self.setEditable(not self.isEditable())
    except Exception:
        pass


@QBaseTable._keymap.bind("Ctrl+X")
@_check_no_viewer
def _(self: QBaseTable):
    self.copyToClipboard(headers=False)
    self.deleteValues()


@QBaseTable._keymap.bind("Ctrl+S")
@_check_no_viewer
def _(self: QBaseTable):
    from tabulous import _io
    from tabulous._qt._history import QtFileHistoryManager

    if path := QtFileHistoryManager.requestPath("w", "Save table"):
        _io.save_file(path, self.getDataFrame())


# (table.show_finder_widget, "Ctrl+F"),
